# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import math
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, TypeVar

from brayns.network import Instance

T = TypeVar("T", bound="Projection")


class Projection(ABC):
    """Base class of all supported camera projections (plugin dependent).

    All camera projections defined in the package inherit from this class.

    Projections can be identified using a unique name (ex: 'perspective').
    """

    @classmethod
    @property
    @abstractmethod
    def name(cls) -> str:
        """Name of the projection to identify it.

        :return: Camera name.
        :rtype: str
        """
        pass

    @abstractmethod
    def look_at(self, height: float) -> float:
        """Focus on a target with given height.

        Update projection properties to look at a target of given height and
        return the minimal distance to see it entirely.

        :param height: Target height.
        :type target: float
        :return: Distance to see target entirely.
        :rtype: float
        """
        pass

    @abstractmethod
    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        pass

    @abstractmethod
    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        pass

    @classmethod
    def from_properties(cls: type[T], message: dict[str, Any]) -> T:
        """Low level API to deserialize from JSON."""
        projection = cls()
        projection.update_properties(message)
        return projection

    def get_properties_with_name(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "name": self.name,
            "params": self.get_properties(),
        }


@dataclass
class OrthographicProjection(Projection):
    """Orthographic camera projection.

    Orthographic camera makes all objects having the same size regardless their
    distance from the camera.

    The viewport width is computed using the aspect ratio of the current
    resolution of the instance (framebuffer size).

    :param height: Viewport height in world coordinates.
    :type height: float
    """

    height: float = 0.0

    @classmethod
    @property
    def name(cls) -> str:
        """Projection name.

        :return: Projection name.
        :rtype: str
        """
        return "orthographic"

    def look_at(self, height: float) -> float:
        """Update viewport height to given one.

        Camera distance doesn't matter in orthographic projections.

        :param height: Target height.
        :type target: float
        :return: Distance to see target entirely.
        :rtype: float
        """
        self.height = height
        return 0

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {"height": self.height}

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.height = message["height"]


class Fovy:
    """Field of view (angle) of a camera."""

    def __init__(self, angle: float, degrees: bool = False) -> None:
        """Construct with angle.

        :param angle: Angle value.
        :type angle: float
        :param degrees: True if angle is in degrees, defaults to False
        :type degrees: bool, optional
        """
        self._angle = math.radians(angle) if degrees else angle

    def __eq__(self, other: object) -> bool:
        """Comparison operator.

        :param other: Object to compare.
        :type other: object
        :return: True if other is Fovy and has same value.
        :rtype: bool
        """
        if not isinstance(other, Fovy):
            return False
        return self._angle == other._angle

    def __hash__(self) -> int:
        """Mark the class as immutable."""
        return hash(self._angle)

    @property
    def radians(self) -> float:
        """Get angle in radians.

        :return: Angle in radians.
        :rtype: float
        """
        return self._angle

    @property
    def degrees(self) -> float:
        """Get angle in degrees.

        :return: Angle in degrees.
        :rtype: float
        """
        return math.degrees(self._angle)

    def get_distance(self, height: float) -> float:
        """Compute the distance to have a viewport with given height.

        :param height: Viewport height.
        :type height: float
        :return: Distance to have the given viewport height.
        :rtype: float
        """
        return height / 2 / math.tan(self.radians / 2)


@dataclass
class PerspectiveProjection(Projection):
    """Perspective camera projection.

    Perspective projection use a field of view angle to compute the size of the
    objects depending on their distance from the camera.

    The field of view (fovy) can be used to compute full screen view of a given
    target object.

    :param fovy: Field of view angle (45 degrees by default).
    :type fovy: Fovy
    :param aperture_radius: Optional aperture radius.
    :type aperture_radius: float
    :param focus_distance: Optional focus distance.
    :type focus_distance: float
    """

    fovy: Fovy = Fovy(45, degrees=True)
    aperture_radius: float = 0.0
    focus_distance: float = 1.0

    @classmethod
    @property
    def name(cls) -> str:
        """Camera name.

        :return: Camera name.
        :rtype: str
        """
        return "perspective"

    def look_at(self, height: float) -> float:
        """Compute camera distance using field of view.

        :param height: Target height.
        :type target: float
        :return: Distance to see target entirely.
        :rtype: float
        """
        return self.fovy.get_distance(height)

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "fovy": self.fovy.degrees,
            "aperture_radius": self.aperture_radius,
            "focus_distance": self.focus_distance,
        }

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.fovy = Fovy(message["fovy"], degrees=True)
        self.aperture_radius = message["aperture_radius"]
        self.focus_distance = message["focus_distance"]


def get_camera_name(instance: Instance) -> str:
    """Retreive the name of the current camera of an instance.

    A camera name is the name of the projection it uses.

    The returned name is the same as Camera.name and can be used to check
    if a given camera is the current one like this:

    :param instance: Instance.
    :type instance: Instance
    :return: Current camera name.
    :rtype: str
    """
    return instance.request("get-camera-type")


def get_camera_projection(instance: Instance, projection_type: type[T]) -> T:
    """Retreive the current camera projection from an instance.

    The provided projection type must be the same as the current one.

    Returned projection is of type ``projection_type``.

    :param instance: Instance.
    :type instance: Instance
    :param projection_type: Projection type (ex: brayns.PerspectiveProjection).
    :type projection_type: type[T]
    :return: Current camera of ``instance``.
    :rtype: T
    """
    name = projection_type.name
    result = instance.request(f"get-camera-{name}")
    return projection_type.from_properties(result)


def set_camera_projection(instance: Instance, camera: Projection) -> None:
    """Set the current camera of the given instance.

    :param instance: Instance.
    :type instance: Instance
    :param camera: Current camera.
    :type camera: Camera
    """
    name = camera.name
    params = camera.get_properties()
    instance.request(f"set-camera-{name}", params)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from .quaternion import Quaternion
from .rotation import Rotation
from .vector import Vector3


@dataclass
class Transform:
    """3D transformation.

    Used to move models in 3D space.

    It is equivalent to a TRS matrix multiplication.

    It means we first scale the point at the origin, then rotate around the same
    origin and finally we translate it to its position.

    :param translation: Translation, defaults to zero.
    :type translation: Vector3, optional
    :param rotation: Rotation, defaults to identity.
    :type rotation: Rotation, optional
    :param scale: Scaling, defaults to one.
    :type scale: Vector3, optional
    """

    translation: Vector3 = Vector3.zero
    rotation: Rotation = Rotation.identity
    scale: Vector3 = Vector3.one

    @staticmethod
    def rotate(rotation: Rotation, center: Vector3 = Vector3.zero) -> Transform:
        """Create a transform to rotate around a given center.

        Rotation around a center is equivalent to a rotation around the origin
        and a translation of center - rotation.apply(center).

        :param rotation: Rotation to apply.
        :type rotation: Rotation
        :param center: Rotation center, defaults to zero.
        :type center: Vector3, optional
        :return: Transform.
        :rtype: Transform
        """
        return Transform(
            translation=center - rotation.apply(center),
            rotation=rotation,
        )

    @classmethod
    @property
    def identity(cls) -> Transform:
        """Create an identity transform (doesn't do anything).

        :return: Identity transform.
        :rtype: Transform
        """
        return Transform()

    def apply(self, value: Vector3) -> Vector3:
        """Apply the transform to the given value.

        :param value: 3D position to transform.
        :type value: Vector3
        :return: Transformed value.
        :rtype: Vector3
        """
        value = self.scale * value
        value = self.rotation.apply(value)
        return value + self.translation


def deserialize_transform(obj: dict[str, Any]) -> Transform:
    return Transform(
        translation=Vector3(*obj["translation"]),
        rotation=Rotation.from_quaternion(Quaternion(*obj["rotation"])),
        scale=Vector3(*obj["scale"]),
    )


def serialize_transform(transform: Transform) -> dict[str, Any]:
    return {
        "translation": list(transform.translation),
        "rotation": list(transform.rotation.quaternion),
        "scale": list(transform.scale),
    }



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import Any


@dataclass
class JsonRpcRequest:
    """Request to send to a running instance of brayns service.

    :param id: Request ID to monitor the request.
        No replies will be received if set to None.
    :type id: int | str | None
    :param method: JSON-RPC method.
    :type method: str
    :param params: Request parameters (usually objects), defaults to None.
    :type params: Any, optional
    :param binary: Request binary data, defaults to empty bytes.
    :type binary: bytes, optional
    """

    id: int | str | None
    method: str
    params: Any = field(default=None, repr=False)
    binary: bytes = field(default=b"", repr=False)


def serialize_request(request: JsonRpcRequest) -> dict[str, Any]:
    message: dict[str, Any] = {
        "jsonrpc": "2.0",
        "method": request.method,
    }
    if request.id is not None:
        message["id"] = request.id
    if request.params is not None:
        message["params"] = request.params
    return message


def serialize_request_to_text(request: JsonRpcRequest) -> str:
    message = serialize_request(request)
    return json.dumps(message, sort_keys=True)


def serialize_request_to_binary(request: JsonRpcRequest) -> bytes:
    json = serialize_request_to_text(request)
    text = json.encode("utf-8")
    json_size = len(text)
    header = json_size.to_bytes(4, byteorder="little", signed=False)
    binary = request.binary
    return b"".join([header, text, binary])



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import math

from .vector import Vector, Vector3


class Quaternion(Vector[float]):
    """Quaternion with XYZW components.

    Provides vector operations (componentwise) and quaternion special operators
    (multiplication, conjugate, etc).

    :param x: X component, defaults to 0.
    :type x: float, optional
    :param y: Y component, defaults to 0.
    :type y: float, optional
    :param z: Z component, defaults to 0.
    :type z: float, optional
    :param w: W component, defaults to 1.
    :type w: float, optional
    """

    @classmethod
    @property
    def identity(cls) -> Quaternion:
        return Quaternion()

    def __new__(
        cls, x: float = 0.0, y: float = 0.0, z: float = 0.0, w: float = 1.0
    ) -> Quaternion:
        return super().__new__(cls, x, y, z, w)

    def __mul__(self, value: int | float | Quaternion) -> Quaternion:
        if isinstance(value, (int, float)):
            return self.unpack(i * value for i in self)
        x0, y0, z0, w0 = self
        x1, y1, z1, w1 = value
        return Quaternion(
            w0 * x1 + x0 * w1 + y0 * z1 - z0 * y1,
            w0 * y1 - x0 * z1 + y0 * w1 + z0 * x1,
            w0 * z1 + x0 * y1 - y0 * x1 + z0 * w1,
            w0 * w1 - x0 * x1 - y0 * y1 - z0 * z1,
        )

    def __rmul__(self, value: int | float | Quaternion) -> Quaternion:
        if isinstance(value, (int, float)):
            return self * value
        return value * self

    def __truediv__(self, value: int | float | Quaternion) -> Quaternion:
        if isinstance(value, (int, float)):
            return self.unpack(i / value for i in self)
        return self * value.inverse

    def __rtruediv__(self, value: int | float | Quaternion) -> Quaternion:
        if isinstance(value, (int, float)):
            return self.unpack(value / i for i in self)
        return value * self.inverse

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]

    @property
    def z(self) -> float:
        return self[2]

    @property
    def w(self) -> float:
        return self[3]

    @property
    def xyz(self) -> Vector3:
        return Vector3(self.x, self.y, self.z)

    @property
    def axis(self) -> Vector3:
        return self.xyz.normalized

    @property
    def angle_radians(self) -> float:
        q = self.normalized
        return 2 * math.acos(q.w)

    @property
    def angle_degrees(self) -> float:
        return math.degrees(self.angle_radians)

    @property
    def conjugate(self) -> Quaternion:
        return Quaternion(-self.x, -self.y, -self.z, self.w)

    @property
    def inverse(self) -> Quaternion:
        return self.conjugate / self.square_norm



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
# Responsible Author: Nadir Roman Guerrero <nadir.romanguerrero@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from brayns.network import Instance


@dataclass
class Framebuffer(ABC):
    """Base class for all framebuffer types.

    Framebuffers define how the rendered image is stored and processed.
    """

    @classmethod
    @property
    @abstractmethod
    def name(cls) -> str:
        """Get the framebuffer name.

        :return: Framebuffer name
        :rtype: str
        """
        pass

    @abstractmethod
    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        pass


@dataclass
class ProgressiveFramebuffer(Framebuffer):
    """Progressive resolution framebuffer.

    :param scale: First-frame reduction factor over original resolution.
    :type scale: int, optional
    """

    scale: int = 4

    @classmethod
    @property
    def name(cls) -> str:
        """Get the framebuffer name.

        :return: Framebuffer name
        :rtype: str
        """
        return "progressive"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {"scale": self.scale}


@dataclass
class StaticFramebuffer(Framebuffer):
    """Static resolution framebuffer."""

    @classmethod
    @property
    def name(cls) -> str:
        """Get the framebuffer name.

        :return: Framebuffer name
        :rtype: str
        """
        return "static"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {}


def set_framebuffer(instance: Instance, framebuffer: Framebuffer) -> None:
    """Stablish the given framebuffer on the engine.

    :param instance: Instance.
    :type instance: Instance
    :param framebuffer: Framebuffer to set on the engine.
    :type framebuffer: Framebuffer
    """
    name = framebuffer.name
    params = framebuffer.get_properties()
    if not params:
        params = None
    instance.request(f"set-framebuffer-{name}", params)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import logging

from .instance import Instance
from .jsonrpc import (
    JsonRpcFuture,
    JsonRpcManager,
    JsonRpcReply,
    JsonRpcRequest,
    serialize_request_to_binary,
    serialize_request_to_text,
)
from .websocket import WebSocket


class Client(Instance):
    def __init__(
        self, websocket: WebSocket, logger: logging.Logger, manager: JsonRpcManager
    ) -> None:
        self._websocket = websocket
        self._logger = logger
        self._manager = manager

    @property
    def connected(self) -> bool:
        return not self._websocket.closed

    def disconnect(self) -> None:
        self._logger.info("Disconnection from Brayns instance.")
        self._websocket.close()
        self._manager.clear()

    def is_running(self, id: int | str) -> bool:
        return self._manager.is_running(id)

    def send(self, request: JsonRpcRequest) -> JsonRpcFuture:
        self._logger.info("Send request: %s.", request)
        self._logger.debug("Request params: %s.", request.params)
        self._logger.info("Request binary: %d bytes.", len(request.binary))
        self._send(request)
        return self._create_future(request.id)

    def poll(self, block: bool) -> None:
        self._logger.debug("Polling messages from Brayns instance.")
        self._websocket.poll(block)

    def cancel(self, id: int | str) -> None:
        self._logger.info("Cancel request with ID %s.", id)
        self.request("cancel", {"id": id})

    def _send(self, request: JsonRpcRequest) -> None:
        if request.binary:
            data = serialize_request_to_binary(request)
            self._logger.debug('Request binary frame data: "%s".', data)
            self._websocket.send_binary(data)
            return
        data = serialize_request_to_text(request)
        self._logger.debug('Request text frame data: "%s".', data)
        self._websocket.send_text(data)

    def _create_future(self, id: int | str | None) -> JsonRpcFuture:
        if id is None:
            reply = JsonRpcReply.for_notifications()
            return JsonRpcFuture.from_reply(reply)
        return JsonRpcFuture(
            task=self._manager.create_task(id),
            cancel=lambda: self.cancel(id),
            poll=lambda block: self.poll(block),
        )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from brayns.network import Instance
from brayns.utils import JsonSchema, deserialize_schema


@dataclass
class Entrypoint:
    """Entrypoint (endpoint) of the JSON-RPC API.

    Describes how a given entrypoint of the JSON-RPC API can be used.

    Available entrypoints can be queried from an instance to inspect the
    Web API of a given instance.

    :param method: JSON-RPC method (ex: 'get-camera-name').
    :type method: str
    :param description: Human readable description.
    :type description: str
    :param plugin: Name of the plugin which loads the entrypoint.
    :type plugin: str
    :param asynchronous: Check wether progress and cancellation are supported.
    :type asynchronous: bool
    :param deprecated: Indicate a removal / renaming in the next major release.
    :type deprecated: bool
    :param params: Schema of input if any, otherwise None.
    :type params: JsonSchema | None
    :param result: Schema of output if any, otherwise None.
    :type result: JsonSchema | None
    """

    method: str
    description: str
    plugin: str
    asynchronous: bool
    deprecated: bool
    params: JsonSchema | None
    result: JsonSchema | None


def get_methods(instance: Instance) -> list[str]:
    """Retreive all JSON-RPC methods from an instance.

    :param instance: Instance to query the methods.
    :type instance: Instance
    :return: List of available methods (depends on plugins loaded).
    :rtype: list[str]
    """
    return instance.request("registry")


def get_entrypoint(instance: Instance, method: str) -> Entrypoint:
    """Retreive an entrypoint using its name (JSON-RPC method).

    :param instance: Instance to query the entrypoint.
    :type instance: Instance
    :param method: JSON-RPC method name.
    :type method: str
    :return: Deserialized entrypoint.
    :rtype: Entrypoint
    """
    params = {"endpoint": method}
    result = instance.request("schema", params)
    return _deserialize_entrypoint(result)


def get_entrypoints(instance: Instance) -> list[Entrypoint]:
    """Retreive all available entrypoints from an instance.

    :param instance: Instance to query the entrypoints.
    :type instance: Instance
    :return: List of available entrypoints (depends on plugins loaded).
    :rtype: list[Entrypoint]
    """
    return [get_entrypoint(instance, method) for method in get_methods(instance)]


def _deserialize_entrypoint(message: dict[str, Any]) -> Entrypoint:
    return Entrypoint(
        method=message["title"],
        description=message["description"],
        plugin=message["plugin"],
        asynchronous=message["async"],
        deprecated=message.get("deprecated", False),
        params=_deserialize_schema(message, "params"),
        result=_deserialize_schema(message, "returns"),
    )


def _deserialize_schema(message: dict[str, Any], key: str) -> JsonSchema | None:
    value = message.get(key)
    if value is None:
        return None
    return deserialize_schema(value)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from brayns.network import Instance
from brayns.utils import Axis, Color3, Vector3

from .model import Model, deserialize_model


@dataclass
class Light(ABC):
    """Base class for all light types.

    :param intensity: Light intensity, defaults to 1.
    :type intensity: float, optional
    :param color: Light color, defaults to white.
    :type color: Color3, optional
    """

    intensity: float = 1.0
    color: Color3 = Color3.white

    @classmethod
    @property
    @abstractmethod
    def name(cls) -> str:
        """Get the light name.

        :return: Light name.
        :rtype: str
        """
        pass

    @abstractmethod
    def get_additional_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        pass

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "intensity": self.intensity,
            "color": list(self.color),
            **self.get_additional_properties(),
        }


@dataclass
class AmbientLight(Light):
    """Ambient light with no particular properties."""

    @classmethod
    @property
    def name(cls) -> str:
        """Get the light name.

        :return: Light name.
        :rtype: str
        """
        return "ambient"

    def get_additional_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {}


@dataclass
class DirectionalLight(Light):
    """Light shining in a given direction.

    :param direction: Light emission direction, defaults to -Z.
    :type direction: Vector3, optional
    """

    direction: Vector3 = Axis.back

    @classmethod
    @property
    def name(cls) -> str:
        """Get the light name.

        :return: Light name.
        :rtype: str
        """
        return "directional"

    def get_additional_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "direction": list(self.direction),
        }


@dataclass
class QuadLight(Light):
    """Rectangular light.

    Emission direction is the positive side (see emission_direction).

    :param bottom_left: Bottom left corner, defaults to origin.
    :type bottom_left: Vector3
    :param edge1: First edge, defaults to +X.
    :type edge1: Vector3, optional
    :param edge2: Second edge, defaults to +Y.
    :type edge2: Vector3, optional
    """

    bottom_left: Vector3 = Vector3.zero
    edge1: Vector3 = Axis.right
    edge2: Vector3 = Axis.up

    @classmethod
    @property
    def name(cls) -> str:
        """Return light name.

        :return: Light name.
        :rtype: str
        """
        return "quad"

    @property
    def emission_direction(self) -> Vector3:
        """Get the emission direction of the light.

        Equal to edge1 x edge2 normalized.

        :return: Emission direction.
        :rtype: Vector3
        """
        return self.edge1.cross(self.edge2).normalized

    def get_additional_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "position": list(self.bottom_left),
            "edge1": list(self.edge1),
            "edge2": list(self.edge2),
        }


def add_light(instance: Instance, light: Light) -> Model:
    """Add a given light to an instance and return its model.

    :param instance: Instance.
    :type instance: Instance
    :param light: Light to add.
    :type light: Light
    :return: Light model.
    :rtype: Model
    """
    name = light.name
    params = light.get_properties()
    result = instance.request(f"add-light-{name}", params)
    return deserialize_model(result)


def clear_lights(instance: Instance) -> None:
    """Remove all lights from an instance.

    :param instance: Instance.
    :type instance: Instance
    """
    instance.request("clear-lights")



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass
from typing import Any, ClassVar

from brayns.core import Fovy, Projection


@dataclass
class CylindricProjection(Projection):
    """Cylindric camera projection used to correct curved screen distorsion.

    :param fovy: Field of view, defaults to OpenDeck's.
    :type fovy: Fovy, optional
    """

    OPENDECK_FOVY: ClassVar[Fovy] = Fovy(48.549, degrees=True)

    fovy: Fovy = OPENDECK_FOVY

    @classmethod
    @property
    def name(cls) -> str:
        """Get projection name.

        :return: Projection name.
        :rtype: str
        """
        return "cylindric"

    def look_at(self, height: float) -> float:
        """Compute camera distance using field of view.

        :param height: Target height.
        :type target: float
        :return: Distance to see target entirely.
        :rtype: float
        """
        return self.fovy.get_distance(height)

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "fovy": self.fovy.degrees,
        }

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.fovy = Fovy(message["fovy"], degrees=True)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import Any


@dataclass
class JsonRpcReply:
    """Reply received from a request sent to a brayns instance.

    :param id: ID of the corresponding request.
    :type id: int | str
    :param result: Reply result (usually objects), defaults to None.
    :type result: Any, optional
    :param binary: Reply binary data, defaults to empty bytes.
    :type binary: bytes, optional
    """

    id: int | str | None
    result: Any = field(default=None, repr=False)
    binary: bytes = field(default=b"", repr=False)

    @staticmethod
    def for_notifications() -> JsonRpcReply:
        return JsonRpcReply(id=None)


def deserialize_reply(message: dict[str, Any]) -> JsonRpcReply:
    return JsonRpcReply(
        id=message["id"],
        result=message["result"],
    )


def deserialize_reply_from_text(data: str) -> JsonRpcReply:
    message = json.loads(data)
    return deserialize_reply(message)


def deserialize_reply_from_binary(data: bytes) -> JsonRpcReply:
    json_size = int.from_bytes(data[0:4], byteorder="little", signed=False)
    text = data[4 : 4 + json_size].decode("utf-8")
    reply = deserialize_reply_from_text(text)
    reply.binary = data[4 + json_size :]
    return reply



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass
from typing import Any

from brayns.version import VERSION

from brayns.network import Instance
from brayns.utils import Error


@dataclass
class VersionError(Error):
    """Error raised when a version mismatch occurs.

    :param local: Local (Python API) version.
    :type local: str
    :param local: Remote (Binary) version.
    :type local: str
    """

    local: str
    remote: str

    def __str__(self) -> str:
        """Format exception.

        :return: Exception description.
        :rtype: str
        """
        local, remote = self.local, self.remote
        return f"Version mismatch with server {local=}, {remote=}"


@dataclass
class Version:
    """Instance version info.

    :param major: Major part.
    :type major: int
    :param minor: Minor part.
    :type minor: int
    :param patch: Patch part.
    :type patch: int
    :param revision: Commit hash.
    :type revision: str
    """

    major: int
    minor: int
    patch: int
    revision: str

    @property
    def release(self) -> tuple[int, int, int]:
        """Return a tuple with major, minor and patch.

        :return: Release tuple.
        :rtype: tuple[int, int, int]
        """
        return (self.major, self.minor, self.patch)

    @property
    def tag(self) -> str:
        """Format version tag 'major.minor.patch'.

        Can be compared to package __version__.

        :return: Version tag.
        :rtype: str
        """
        return ".".join(str(i) for i in self.release)

    def check(self, local: str = VERSION) -> None:
        """Check that self is compatible with local version.

        :param local: API version, defaults to VERSION.
        :type local: str, optional
        :raises VersionError: Version mismatch.
        """
        remote = self.tag
        if remote != local:
            raise VersionError(local, remote)


def get_version(instance: Instance) -> Version:
    """Retreive the version of a backend instance.

    :param instance: Instance.
    :type instance: Instance
    :return: Version info.
    :rtype: Version
    """
    result = instance.request("get-version")
    return _deserialize_version(result)


def check_version(instance: Instance) -> None:
    """Check that the version of the instance is compatible with the API.

    :param instance: Instance to check.
    :type instance: Instance
    """
    version = get_version(instance)
    version.check()


def _deserialize_version(message: dict[str, Any]) -> Version:
    return Version(
        major=message["major"],
        minor=message["minor"],
        patch=message["patch"],
        revision=message["revision"],
    )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Plugins subpackage.

It includes all science related features and optional components.
"""

from .atlas import (
    AtlasColumnHighlight,
    AtlasDensity,
    AtlasDistanceType,
    AtlasFlatmapAreas,
    AtlasLayerDistance,
    AtlasOrientationField,
    AtlasShellOutline,
    AtlasUsecase,
    ColumnNeighbor,
    ColumnPosition,
    get_atlas_usecases,
)
from .bbp import BbpCells, BbpLoader, BbpReport, BbpReportType
from .cell_placement import CellPlacementLoader
from .circuit import CircuitColorMethod, get_circuit_ids, set_circuit_thickness
from .cylindric_camera import CylindricProjection
from .dti import DtiLoader
from .morphology import (
    GeometryType,
    Morphology,
    MorphologyColorMethod,
    MorphologyLoader,
)
from .nrrd import NrrdLoader, VoxelType
from .protein import ProteinColorScheme, ProteinLoader
from .sonata import (
    SonataEdgePopulation,
    SonataLoader,
    SonataNodePopulation,
    SonataNodes,
    SonataReport,
    SonataReportType,
)
from .xyz import XyzLoader

__all__ = [
    "AtlasColumnHighlight",
    "AtlasDensity",
    "AtlasDistanceType",
    "AtlasFlatmapAreas",
    "AtlasLayerDistance",
    "AtlasOrientationField",
    "AtlasShellOutline",
    "AtlasUsecase",
    "BbpCells",
    "BbpLoader",
    "BbpReport",
    "BbpReportType",
    "CellPlacementLoader",
    "CircuitColorMethod",
    "ColumnNeighbor",
    "ColumnPosition",
    "CylindricProjection",
    "DtiLoader",
    "GeometryType",
    "get_atlas_usecases",
    "get_circuit_ids",
    "Morphology",
    "MorphologyColorMethod",
    "MorphologyLoader",
    "NrrdLoader",
    "ProteinColorScheme",
    "ProteinLoader",
    "set_circuit_thickness",
    "SonataEdgePopulation",
    "SonataLoader",
    "SonataNodePopulation",
    "SonataNodes",
    "SonataReport",
    "SonataReportType",
    "VoxelType",
    "XyzLoader",
]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from typing import NamedTuple

from brayns.network import Connector, Instance

from .process import Process
from .service import Service


class Manager(NamedTuple):
    """Wrapper to manage a ``Process`` connected to an ``Instance``.

    Contains a process running braynService and an instance connected to it.

    Once done with this object, ``stop()`` must be called to disconnect the
    instance and kill the process, using the context manager or by hand.

    :param process: Process running braynsService.
    :type process: Process
    :param instance: Instance connected to ``process``.
    :type instance: Instance
    """

    process: Process
    instance: Instance

    def __enter__(self) -> Manager:
        """Context manager enter, just return self."""
        return self

    def __exit__(self, *_) -> None:
        """Call stop when exiting the context manager."""
        self.stop()

    def stop(self) -> None:
        """Disconnect the instance and kill the process."""
        self.instance.disconnect()
        self.process.stop()


def start(service: Service, connector: Connector) -> Manager:
    """Start a braynsService instance and connect to it.

    Return the service process and the connected instance (see ``Manager``).

    :param service: Service specifications.
    :type service: Service
    :param connector: Connection specifications.
    :type connector: Connector
    :return: Running process and connected instance.
    :rtype: Manager
    """
    process = service.start()
    try:
        instance = connector.connect()
    except BaseException:
        process.stop()
        raise
    return Manager(process, instance)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from brayns.network import Instance
from brayns.utils import Vector2, Vector3


@dataclass
class PickResult:
    """Information about the model found at an inspected screen position.

    Metadata depend on the kind of model found at given position and are
    specific to the primitive (subpart) hitted.

    :param position: World position matching screen coordinates.
    :type position: Vector3
    :param model_id: ID of them model at screen position.
    :type model_id: int
    :param metadata: Information about the model primitive at given position.
    :type metadata: Any
    """

    position: Vector3
    model_id: int
    metadata: Any


def pick(instance: Instance, position: Vector2) -> PickResult | None:
    """Pick a given screen position and inspect it.

    Screen position must be normalized with [0, 0] being bottom-left of
    and [1, 1] top-right.

    If a model is found on given coordinates, return its ID with the world
    position corresponding to the pixel where the model was rendered.

    If no models are found, then None is returned.

    :param instance: Instance to inspect.
    :type instance: Instance
    :param x: Screen position X normalized.
    :type x: float
    :return: Information about model in given XY if any, otherwise None.
    :rtype: PickResult | None
    """
    params = {"position": list(position)}
    result = instance.request("inspect", params)
    return _deserialize_pick_result(result)


def _deserialize_pick_result(message: dict[str, Any]) -> PickResult | None:
    hit = message["hit"]
    if not hit:
        return None
    return PickResult(
        position=Vector3(*message["position"]),
        model_id=message["model_id"],
        metadata=message["metadata"],
    )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import logging
import time
from dataclasses import dataclass, field
from typing import cast

from .client import Client
from .instance import Instance
from .jsonrpc import JsonRpcManager
from .listener import Listener
from .logger import Logger
from .websocket import (
    ServiceUnavailableError,
    SslClientContext,
    WebSocketClient,
    WebSocketConnector,
)


@dataclass
class Connector:
    """Used to connect to a braynsService instance.

    A connector must at least have the URI the braynsService instance was
    started with.

    URI is always in format host:port, wss:// or ws:// will be added depending
    if an SSL context is provided (SSL is disabled if context is None).

    SSL context can be provided for a secure connection. If the instance uses a
    certificate signed by a CA installed on the local machine, the default value
    of brayns.SslClientContext() can be used. Otherwise, the CA file or path
    must be provided in the SSL context constructor.

    Binary messages received from the instance are not JSON-RPC requests but can
    be handled using an optional callback (see ``binary_handler``).

    If you don't know when your instance of braynsService will be ready when you
    call connect(), you can set ``max_attempts`` to None to try to connect in
    loop until it works (or with a maximum count and a delay to have a timeout).

    :param uri: Instance URI with format 'host:port'.
    :type uri: str
    :param ssl_context: SSL context if secure, defaults to None.
    :type ssl_context: SslClientContext | None, optional
    :param logger: Instance logger, defaults to brayns.Logger().
    :type logger: logging.Logger, optional
    :param max_attempts: Max connection attempts, defaults to 1.
    :type max_attempts: int | None, optional
    :param attempt_period: Delay in seconds between attempts, defaults to 0.1.
    :type attempt_period: float, optional
    """

    uri: str
    ssl_context: SslClientContext | None = None
    logger: logging.Logger = field(default_factory=Logger)
    max_attempts: int | None = 1
    attempt_period: float = 0.1

    def connect(self) -> Instance:
        """Connect to instance and return it.

        Try to connect ``max_attempts`` times waiting ``attempt_period`` between
        two attempts. If it fails, ServiceUnavailableError will be raised.

        :raises WebSocketError
        :return: Connected braynsService instance.
        :rtype: Instance
        """
        manager = JsonRpcManager(self.logger)
        listener = Listener(self.logger, manager)
        connector = WebSocketConnector(self.uri, listener, self.ssl_context)
        websocket = self._open_websocket(connector)
        return Client(websocket, self.logger, manager)

    def _open_websocket(self, connector: WebSocketConnector) -> WebSocketClient:
        self.logger.info('Connection to instance at "%s".', self.uri)
        try:
            websocket = self._try_connect(connector)
        except Exception as e:
            self.logger.info('Connection failed: "%s".', e)
            raise
        self.logger.info("Successfully connected.")
        return websocket

    def _try_connect(self, connector: WebSocketConnector) -> WebSocketClient:
        if self.max_attempts is None:
            return self._try_connect_forever(connector)
        return self._try_connect_with_max_attempts(connector)

    def _try_connect_forever(self, connector: WebSocketConnector) -> WebSocketClient:
        while True:
            self.logger.debug("Connection attempt.")
            try:
                return connector.connect()
            except ServiceUnavailableError:
                time.sleep(self.attempt_period)

    def _try_connect_with_max_attempts(
        self, connector: WebSocketConnector
    ) -> WebSocketClient:
        max_attempts = cast(int, self.max_attempts)
        count = 1
        while True:
            self.logger.debug("Connection attempt %d/%d.", count, max_attempts)
            try:
                return connector.connect()
            except ServiceUnavailableError:
                if count == max_attempts:
                    self.logger.info("Max attempts reached.")
                    raise
                time.sleep(self.attempt_period)
            count += 1



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from typing import ClassVar

from brayns.network import Instance


class CircuitColorMethod:
    """Available coloring methods for circuits."""

    ID: ClassVar[str] = "id"
    ETYPE: ClassVar[str] = "etype"
    MTYPE: ClassVar[str] = "mtype"
    LAYER: ClassVar[str] = "layer"
    REGION: ClassVar[str] = "region"
    HEMISPHERE: ClassVar[str] = "hemisphere"
    MORPHOLOGY: ClassVar[str] = "morphology"
    MORPHOLOGY_CLASS: ClassVar[str] = "morphology class"
    MORPHOLOGY_SECTION: ClassVar[str] = "morphology section"
    SUBREGION: ClassVar[str] = "subregion"
    SYNAPSE_CLASS: ClassVar[str] = "synapse class"


def get_circuit_ids(instance: Instance, model_id: int) -> list[int]:
    """Retreive the list of loaded neurons / astrocytes / vasculatures by ID.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: ID of the circuit model.
    :type model_id: int
    :return: ID list of the element loaded in given circuit.
    :rtype: list[int]
    """
    params = {"model_id": model_id}
    return instance.request("get-circuit-ids", params)


def set_circuit_thickness(
    instance: Instance, model_id: int, radius_multiplier: float
) -> None:
    """Multiply the radius of all primitives of a circuit by given factor.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: Circuit model ID.
    :type model_id: int
    :param radius_multiplier: Scaling factor for capsules and spheres.
    :type radius_multiplier: float
    """
    params = {
        "model_id": model_id,
        "radius_multiplier": radius_multiplier,
    }
    instance.request("set-circuit-thickness", params)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any


class JsonType(Enum):
    """Enumeration of available JSON types.

    The value is the string from the JSON schema standard.

    :param UNDEFINED: Any type is allowed.
    :param NULL: No types are allowed.
    :param BOOLEAN: Boolean.
    :param INTEGER: Integral number.
    :param NUMBER: Any number.
    :param STRING: String.
    :param ARRAY: Array.
    :param OBJECT: Object.
    """

    UNDEFINED = "undefined"
    NULL = "null"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    NUMBER = "number"
    STRING = "string"
    ARRAY = "array"
    OBJECT = "object"


@dataclass
class JsonSchema:
    """JSON schema describing JSON-RPC message format.

    :param title: Title (class name).
    :type title: str
    :param description: Human readable description.
    :type description: str
    :param type: JSON type (null, boolean, number, string, object, ...).
    :type type: JsonType
    :param read_only: If true, this field cannot be assigned.
    :type read_only: bool
    :param write_only: If true, this field cannot get retreived.
    :type write_only: bool
    :param default: Optional default value.
    :type default: Any
    :param minimum: Optional minimum value for numbers.
    :type minimum: float | None
    :param maximum: Optional maximum value for numbers.
    :type maximum: float | None
    :param items: Optional item schema value for arrays.
    :type items: JsonSchema | None
    :param min_items: Optional minimum item count for arrays.
    :type min_items: int | None
    :param max_items: Optional maximum item count for arrays.
    :type max_items: int | None
    :param properties: Optional fixed properties schema for objects.
    :type properties: dict[str, JsonSchema]
    :param required: Optional list of required properties for objects.
    :type required: list[str]
    :param additional_properties: Schema of additional properties for objects.
        If False then no additional properties are allowed (only fixed).
        If None then any additional properties are allowed (usually ignored).
    :type additional_properties: bool | JsonSchema | None
    :param one_of: If not empty, the schema is a union of several schemas.
    :type one_of: list[JsonSchema]
    :param enum: If not empty, the message must be in these values.
    :type enum: list[Any]
    """

    title: str = ""
    description: str = ""
    type: JsonType = JsonType.UNDEFINED
    read_only: bool = False
    write_only: bool = False
    default: Any = None
    minimum: float | None = None
    maximum: float | None = None
    items: JsonSchema | None = None
    min_items: int | None = None
    max_items: int | None = None
    properties: dict[str, JsonSchema] = field(default_factory=dict)
    required: list[str] = field(default_factory=list)
    additional_properties: bool | JsonSchema | None = None
    one_of: list[JsonSchema] = field(default_factory=list)
    enum: list[Any] = field(default_factory=list)


def deserialize_schema(message: dict[str, Any]) -> JsonSchema:
    """Deserialize dictionary to JSON schema.

    :param message: Parsed JSON object.
    :type message: dict[str, Any]
    :return: JSON schema with supported fields.
    :rtype: JsonSchema
    """
    return JsonSchema(
        title=message.get("title", ""),
        description=message.get("description", ""),
        type=_deserialize_type(message),
        read_only=message.get("readOnly", False),
        write_only=message.get("writeOnly", False),
        default=message.get("default"),
        minimum=message.get("minimum"),
        maximum=message.get("maximum"),
        items=_deserialize_items(message),
        min_items=message.get("minItems"),
        max_items=message.get("maxItems"),
        properties=_deserialize_properties(message),
        required=message.get("required", []),
        additional_properties=_deserialize_additional(message),
        one_of=_deserialize_one_of(message),
        enum=message.get("enum", []),
    )


def serialize_schema(schema: JsonSchema) -> dict[str, Any]:
    """Serialize JSON schema to dictionary.

    :param schema: JSON schema.
    :type schema: JsonSchema
    :return: Dictionary that can be stringified to raw JSON.
    :rtype: dict[str, Any]
    """
    message = dict[str, Any]()
    if schema.title:
        message["title"] = schema.title
    if schema.description:
        message["description"] = schema.description
    if schema.type is not JsonType.UNDEFINED:
        message["type"] = schema.type.value
    if schema.read_only:
        message["readOnly"] = schema.read_only
    if schema.write_only:
        message["writeOnly"] = schema.write_only
    if schema.default is not None:
        message["default"] = schema.default
    if schema.minimum is not None:
        message["minimum"] = schema.minimum
    if schema.maximum is not None:
        message["maximum"] = schema.maximum
    if schema.items is not None:
        message["items"] = serialize_schema(schema.items)
    if schema.min_items is not None:
        message["minItems"] = schema.min_items
    if schema.max_items is not None:
        message["maxItems"] = schema.max_items
    if schema.properties:
        message["properties"] = _serialize_properties(schema)
    if schema.required:
        message["required"] = schema.required
    if schema.additional_properties is not None:
        message["additionalProperties"] = _serialize_additional(schema)
    if schema.one_of:
        message["oneOf"] = _serialize_one_of(schema)
    if schema.enum:
        message["enum"] = schema.enum
    return message


def _deserialize_type(obj: dict[str, Any]) -> JsonType:
    value = obj.get("type")
    if value is None:
        return JsonType.UNDEFINED
    return JsonType(value)


def _deserialize_items(obj: dict[str, Any]) -> JsonSchema | None:
    value = obj.get("items")
    if value is None:
        return None
    if isinstance(value, dict):
        return deserialize_schema(value)
    raise TypeError(f"JSON API type error {value}")


def _deserialize_properties(obj: dict[str, Any]) -> dict[str, JsonSchema]:
    value = obj.get("properties", dict[str, Any]())
    return {key: deserialize_schema(schema) for key, schema in value.items()}


def _deserialize_additional(obj: dict[str, Any]) -> bool | JsonSchema | None:
    value = obj.get("additionalProperties")
    if value is None:
        return None
    if value is False:
        return value
    if isinstance(value, dict):
        return deserialize_schema(value)
    raise TypeError(f"JSON API type error {value}")


def _deserialize_one_of(obj: dict[str, Any]) -> list[JsonSchema]:
    value = obj.get("oneOf", list[dict[str, Any]]())
    return [deserialize_schema(one_of) for one_of in value]


def _serialize_properties(schema: JsonSchema) -> dict[str, Any]:
    return {key: serialize_schema(value) for key, value in schema.properties.items()}


def _serialize_additional(schema: JsonSchema) -> bool | dict[str, Any] | None:
    properties = schema.additional_properties
    if properties is None:
        return None
    if isinstance(properties, bool):
        return properties
    return serialize_schema(properties)


def _serialize_one_of(schema: JsonSchema) -> list[dict[str, Any]]:
    return [serialize_schema(one_of) for one_of in schema.one_of]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Any

from brayns.core import Model, deserialize_model
from brayns.network import Instance
from brayns.utils import Color4, JsonSchema, Vector, deserialize_schema


class AtlasUsecase(ABC):
    """Base class for all atlas visualization usecases."""

    @classmethod
    @property
    @abstractmethod
    def name(cls) -> str:
        """Low level API."""
        pass

    @abstractmethod
    def get_properties(self) -> dict[str, Any]:
        """Low level API."""
        pass

    def run(self, instance: Instance, model_id: int) -> Model:
        """Create a model to visualize the usecase on an atlas volume.

        :param instance: Instance.
        :type instance: Instance
        :param model_id: ID of the atlas volume model.
        :type model_id: int
        :return: Model created for the usecase.
        :rtype: Model
        """
        params = {
            "model_id": model_id,
            "use_case": self.name,
            "params": self.get_properties(),
        }
        result = instance.request("visualize-atlas-usecase", params)
        return deserialize_model(result)


class AtlasDensity(AtlasUsecase):
    """Display the volume density.

    The relation between the density and the color can be changed using the
    color ramp of the model created.
    """

    @classmethod
    @property
    def name(cls) -> str:
        return "Density"

    def get_properties(self) -> dict[str, Any]:
        return {}


class AtlasFlatmapAreas(AtlasUsecase):
    """Display the volumne flatmap areas."""

    @classmethod
    @property
    def name(cls) -> str:
        return "Flatmap areas"

    def get_properties(self) -> dict[str, Any]:
        return {}


class ColumnPosition(Vector[int]):
    """Column position XZ.

    :param x: X component.
    :type x: float
    :param z: Z component.
    :type z: float
    """

    def __new__(cls, x: int = 0, z: int = 0) -> ColumnPosition:
        return super().__new__(cls, x, z)

    @property
    def x(self) -> float:
        return self[0]

    @property
    def z(self) -> float:
        return self[1]


@dataclass
class ColumnNeighbor:
    """Column neighbor position and color.

    The position is relative to the main column coordinates.
    """

    position: ColumnPosition
    color: Color4


@dataclass
class AtlasColumnHighlight(AtlasUsecase):
    """Place a column and (optionally) its neighbors.

    Columns are always aligned on the Y axis so their position uses only XZ.

    :param position: Column coordinates.
    :type position: ColumnPosition
    :param color: Color to apply to the column.
    :type color: Color4
    :param neighbors: Neighbors to color, defaults to empty.
    :type neighbors: list[ColumnNeighbor], optional
    """

    position: ColumnPosition
    color: Color4
    neighbors: list[ColumnNeighbor] = field(default_factory=list)

    @classmethod
    @property
    def name(cls) -> str:
        return "Highlight columns"

    def get_properties(self) -> dict[str, Any]:
        return {
            "xz_coordinate": list(self.position),
            "color": list(self.color),
            "neighbours": [
                {
                    "relative_xz": list(neighbor.position),
                    "color": list(neighbor.color),
                }
                for neighbor in self.neighbors
            ],
        }


class AtlasDistanceType(Enum):
    """Type of distance displayed for an atlas layer.

    :params LOWER: Distance of the current layer from the one below.
    :params UPPER: Distance of the current layer from the one above.
    """

    LOWER = "lower"
    UPPER = "upper"


@dataclass
class AtlasLayerDistance(AtlasUsecase):
    """Display the volume as the distance of a layer from its neighbors."""

    distance_type: AtlasDistanceType

    @classmethod
    @property
    def name(cls) -> str:
        return "Layer distance"

    def get_properties(self) -> dict[str, Any]:
        return {"type": self.distance_type.value}


class AtlasOrientationField(AtlasUsecase):
    """Display the volume data as an orientation field."""

    @classmethod
    @property
    def name(cls) -> str:
        return "Orientation field"

    def get_properties(self) -> dict[str, Any]:
        return {}


class AtlasShellOutline(AtlasUsecase):
    """Display the volume using its shell outline.

    This is the default atlas volume usecase and is always supported.
    """

    @classmethod
    @property
    def name(cls) -> str:
        return "Outline mesh shell"

    def get_properties(self) -> dict[str, Any]:
        return {}


def get_atlas_usecases(
    instance: Instance, model_id: int
) -> list[tuple[str, JsonSchema]]:
    """Return the list of visualization usecases supported by an atlas volume.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: ID of the atlas volume model.
    :type model_id: int
    :return: Names of the supported usecase (see ``AtlasUsecase.name``)
    :rtype: list[str]
    """
    params = {"model_id": model_id}
    result: list[dict[str, Any]] = instance.request(
        "get-available-atlas-usecases", params
    )
    return [
        (item["name"], deserialize_schema(item["params_schema"])) for item in result
    ]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.

# Responsible Authors:
#   Adrien Christian Fleury <adrien.fleury@epfl.ch>
#   Nadir Roman Guerrero <nadir.romanguerrero@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any

from brayns.network import Future, Instance, JsonRpcFuture, JsonRpcReply
from brayns.utils import Resolution

from .camera import Camera
from .renderer import Renderer
from .view import serialize_view


class GBufferChannel(Enum):
    """All available channels to export."""

    COLOR = "color"
    DEPTH = "depth"
    ALBEDO = "albedo"
    NORMAL = "normal"


@dataclass
class GBufferExporter:
    """Helper class to export the G-Buffers of an instance as an EXR encoded
    image.

    A temporary context will be created in the instance to avoid changing
    its state.

    For None parameters, the current values of the instance are used.

    :param channels: List of g-buffer channels to export.
    :type channels: list[GBufferChannel]
    :param resolution: Image resolution, defaults to None.
    :type resolution: Resolution | None, optional
    :param camera: Camera used to render, defaults to None.
    :type camera: Camera | None, optional
    :param renderer: Renderer used to render, defaults to None.
    :type renderer: Renderer | None, optional
    :param frame: Simulation index, defaults to None.
    :type frame: int | None, optional
    """

    channels: list[GBufferChannel]
    resolution: Resolution | None = None
    camera: Camera | None = None
    renderer: Renderer | None = None
    frame: int | None = None

    def save(self, instance: Instance, path: str) -> None:
        """Download and save the g-buffers locally under given file.

        Path is on the local machine (running current script).

        The ".exr" extension will be appended automatically if not present.

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        """
        task = self.save_task(instance, path)
        task.wait_for_result()

    def save_task(self, instance: Instance, path: str) -> Future[None]:
        """Asynchronous version of ``save``.

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        :return: Future to monitor the task.
        :rtype: Future[None]
        """

        if Path(path).suffix != ".exr":
            path = path + ".exr"

        task = self._download(instance)
        return Future(task, lambda reply: _save_exr(reply, path))

    def save_remotely(self, instance: Instance, path: str) -> None:
        """Exports the g-buffers remotely under given file.

        Path is on the remote machine (running instance backend).

        The ".exr" extension will be appended automatically if not present.

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        """
        task = self.save_remotely_task(instance, path)
        task.wait_for_result()

    def save_remotely_task(self, instance: Instance, path: str) -> Future[None]:
        """Asynchronous version of ``save_remotely``.

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        :return: Future to monitor the task.
        :rtype: Future[None]
        """
        params = _serialize_export(self, path)
        task = _task(instance, params)
        return Future(task, lambda _: None)

    def download(self, instance: Instance) -> bytes:
        """Download the exported g-buffers as EXR encoded bytes.

        :param instance: Instance.
        :type instance: Instance
        :return: Image data.
        :rtype: bytes
        """
        task = self.download_task(instance)
        return task.wait_for_result()

    def download_task(self, instance: Instance) -> Future[bytes]:
        """Asynchronous version of ``download``.

        :param instance: Instance.
        :type instance: Instance
        :return: Future to monitor the task.
        :rtype: Future[bytes]
        """
        task = self._download(instance)
        return Future(task, lambda reply: reply.binary)

    def _download(self, instance: Instance) -> JsonRpcFuture:
        params = _serialize_export(self)
        return _task(instance, params)


def _serialize_export(
    export: GBufferExporter, path: str | None = None
) -> dict[str, Any]:
    message: dict[str, Any] = {
        "channels": [channel.value for channel in export.channels]
    }

    if path is not None:
        message["file_path"] = path
    if export.resolution is not None:
        message["resolution"] = list(export.resolution)
    if export.camera is not None:
        camera = export.camera
        message["camera_view"] = serialize_view(camera.view)
        message["camera"] = camera.projection.get_properties_with_name()
    if export.renderer is not None:
        message["renderer"] = export.renderer.get_properties_with_name()
    if export.frame is not None:
        message["simulation_frame"] = export.frame

    return message


def _task(instance: Instance, params: dict[str, Any]) -> JsonRpcFuture:
    return instance.task("export-gbuffers", params)


def _save_exr(reply: JsonRpcReply, path: str) -> None:
    data = reply.binary
    with open(path, "wb") as file:
        file.write(data)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
# Responsible Author: Nadir Roman Guerrero <nadir.romanguerrero@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from brayns.core import Loader

from .morphology import Morphology, serialize_morphology


@dataclass
class CellPlacementLoader(Loader):
    """circuit.morhpologies.h5 file loader.

    Loads morphology circuits from the circuit builder intermediate files.

    :param morphologies_folder: Path to the folder containing the morphologies.
    :type morphologies_folder: str
    :param density: Density of morphologies to load [0-1].
    :type density: float
    :param extension: Optional morphology file extension to load.
    :type extension: str | None, optional
    :param ids: IDs of the nodes to load, overrides density if not None.
    :type ids: list[str] | None, optional
    :param morphology: Morphology loading settings.
    :type morphology: Morphology, optional
    """

    morphologies_folder: str
    density: float = 1.0
    extension: str | None = None
    ids: list[int] | None = None
    morphology: Morphology = field(default_factory=Morphology)

    @classmethod
    @property
    def name(cls) -> str:
        return "Cell placement loader"

    def get_properties(self) -> dict[str, Any]:
        message: dict[str, Any] = {
            "morphology_folder": self.morphologies_folder,
            "morphology_parameters": serialize_morphology(self.morphology),
        }
        if self.extension is not None:
            message["extension"] = self.extension
        if self.density != 1.0:
            message["percentage"] = self.density
        if self.ids is not None:
            message["ids"] = self.ids
        return message



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
# Responsible Author: Nadir Roman Guerrero <nadir.romanguerrero@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass
from enum import Enum
from typing import Any, ClassVar

from brayns.core import Loader


class VoxelType(Enum):
    """Defines the representation of the voxels of an NRRD volume.

    :param SCALAR: Each voxel holds a single scalar value.
    :param ORIENTATION: Each voxel holds a quaternion (orientation field).
    :param FLATMAP: Each voxel is a pair of integer.
    """

    SCALAR = "scalar"
    ORIENTATION = "orientation"
    FLATMAP = "flatmap"


@dataclass
class NrrdLoader(Loader):
    """Loader for NRRD volumes.

    :param voxel_type: Defines the representation of the volume voxels.
    :type voxel_type: VoxelType
    """

    NRRD: ClassVar[str] = "nrrd"

    voxel_type: VoxelType

    @classmethod
    @property
    def name(cls) -> str:
        return "NRRD loader"

    def get_properties(self) -> dict[str, Any]:
        return {"type": self.voxel_type.value}



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from .transform import Rotation
from .vector import Vector3, componentwise_max, componentwise_min


@dataclass
class Bounds:
    """Axis aligned bounding box.

    :param min: Minimum XYZ.
    :type min: Vector3
    :param max: Maximum XYZ.
    :type max: Vector3
    """

    min: Vector3
    max: Vector3

    @staticmethod
    def of(values: list[Vector3]) -> Bounds:
        """Compute the bounds of given 3D points.

        :param values: List of points to include in bounds.
        :type values: list[Vector3]
        :return: Points boundary
        :rtype: Bounds
        """
        return Bounds(
            min=componentwise_min(values),
            max=componentwise_max(values),
        )

    @classmethod
    @property
    def empty(cls) -> Bounds:
        """Create empty bounds.

        :return: Empty bounds in [0, 0, 0].
        :rtype: Bounds
        """
        return Bounds(Vector3.zero, Vector3.zero)

    @classmethod
    @property
    def unit(cls) -> Bounds:
        """Create unit bounds (size = [1, 1, 1]).

        :return: Unit bounds centered in [0, 0, 0].
        :rtype: Bounds
        """
        return Bounds(-Vector3.one / 2, Vector3.one / 2)

    @property
    def center(self) -> Vector3:
        """Compute the center point of the bounds.

        :return: 3D point of the center.
        :rtype: Vector3
        """
        return (self.min + self.max) / 2

    @property
    def size(self) -> Vector3:
        """Compute the size of the bounds.

        :return: Size XYZ (width, height, depth).
        :rtype: Vector3
        """
        return self.max - self.min

    @property
    def width(self) -> float:
        """Compute the width of the bounds.

        :return: size.x.
        :rtype: float
        """
        return self.size.x

    @property
    def height(self) -> float:
        """Compute the height of the bounds.

        :return: size.y.
        :rtype: float
        """
        return self.size.y

    @property
    def depth(self) -> float:
        """Compute the depth of the bounds.

        :return: size.z.
        :rtype: float
        """
        return self.size.z

    @property
    def corners(self) -> list[Vector3]:
        """List the 8 corners of the box.

        :return: List of corner points.
        :rtype: list[Vector3]
        """
        return [
            self.min,
            Vector3(self.min.x, self.min.y, self.max.z),
            Vector3(self.min.x, self.max.y, self.min.z),
            Vector3(self.min.x, self.max.y, self.max.z),
            Vector3(self.max.x, self.min.y, self.min.z),
            Vector3(self.max.x, self.min.y, self.max.z),
            Vector3(self.max.x, self.max.y, self.min.z),
            self.max,
        ]

    def translate(self, translation: Vector3) -> Bounds:
        """Translate bounds by given value.

        :param translation: Translation for min and max.
        :type translation: Vector3
        :return: Translated bounds.
        :rtype: Bounds
        """
        return Bounds(
            min=self.min + translation,
            max=self.max + translation,
        )

    def rotate(self, rotation: Rotation, center: Vector3 = Vector3.zero) -> Bounds:
        """Rotate bounds by given value around optional center.

        :param rotation: Rotation to apply.
        :type rotation: Rotation
        :param center: Rotation center, defaults to Vector3.zero.
        :type center: Vector3, optional
        :return: New axis aligned bounds after rotation.
        :rtype: Bounds
        """
        return Bounds.of([rotation.apply(corner, center) for corner in self.corners])

    def rescale(self, scale: Vector3) -> Bounds:
        """Multiply limits by given scale componentwise.

        :param scale: Scale XYZ.
        :type scale: Vector3
        :return: Rescaled bounds.
        :rtype: Bounds
        """
        return Bounds(
            min=scale * self.min,
            max=scale * self.max,
        )


def merge_bounds(values: list[Bounds]) -> Bounds:
    """Compute the union of all given bounds.

    Returns Bounds.empty if values are empty.

    Assume that all bounds are valid (ie min <= max for each component).

    :param values: Bounds to merge.
    :type values: list[Bounds]
    :return: Union of all bounds in values.
    :rtype: Bounds
    """
    return Bounds(
        min=componentwise_min([value.min for value in values]),
        max=componentwise_max([value.max for value in values]),
    )


def deserialize_bounds(obj: dict[str, Any]) -> Bounds:
    return Bounds(
        min=Vector3(*obj["min"]),
        max=Vector3(*obj["max"]),
    )


def serialize_bounds(bounds: Bounds) -> dict[str, Any]:
    return {
        "min": list(bounds.min),
        "max": list(bounds.max),
    }



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import asyncio
import ssl
from dataclasses import dataclass

from websockets.client import connect
from websockets.exceptions import WebSocketException

from .async_web_socket import AsyncWebSocket
from .errors import (
    InvalidServerCertificateError,
    ProtocolError,
    ServiceUnavailableError,
)
from .event_loop import EventLoop
from .web_socket_client import WebSocketClient
from .web_socket_listener import WebSocketListener


@dataclass
class SslClientContext:
    cafile: str | None = None
    capath: str | None = None
    cadata: str | None = None


@dataclass
class WebSocketConnector:
    uri: str
    listener: WebSocketListener
    ssl_context: SslClientContext | None = None

    def connect(self) -> WebSocketClient:
        loop = EventLoop()
        try:
            websocket = loop.run(_connect(self.uri, self.ssl_context)).result()
        except BaseException:
            loop.close()
            raise
        return WebSocketClient(websocket, loop, self.listener)


async def _connect(uri: str, context: SslClientContext | None) -> AsyncWebSocket:
    try:
        websocket = await connect(
            uri=_format_uri(uri, context),
            ssl=_try_create_ssl_context(context),
            ping_interval=None,
            close_timeout=0,
            max_size=int(2e9),
        )
        return AsyncWebSocket(websocket)
    except WebSocketException as e:
        raise ProtocolError(str(e))
    except ssl.SSLError as e:
        raise InvalidServerCertificateError(str(e))
    except OSError as e:
        raise ServiceUnavailableError(str(e))
    except asyncio.TimeoutError as e:
        raise ServiceUnavailableError(str(e))


def _format_uri(uri: str, context: SslClientContext | None) -> str:
    protocol = "ws://" if context is None else "wss://"
    return protocol + uri


def _try_create_ssl_context(context: SslClientContext | None) -> ssl.SSLContext | None:
    if context is None:
        return None
    return ssl.create_default_context(
        cafile=context.cafile, capath=context.capath, cadata=context.cadata
    )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from collections import deque

from .json_rpc_error import JsonRpcError
from .json_rpc_progress import JsonRpcProgress
from .json_rpc_reply import JsonRpcReply


class JsonRpcTask:
    @staticmethod
    def from_reply(reply: JsonRpcReply) -> JsonRpcTask:
        task = JsonRpcTask()
        task.set_reply(reply)
        return task

    def __init__(self) -> None:
        self._reply: JsonRpcReply | None = None
        self._error: JsonRpcError | None = None
        self._progresses = deque[JsonRpcProgress]()

    def is_ready(self) -> bool:
        return self._reply is not None or self._error is not None

    def has_progress(self) -> bool:
        return bool(self._progresses)

    def get_reply(self) -> JsonRpcReply:
        if self._error is not None:
            raise self._error
        if self._reply is None:
            raise RuntimeError("Task is still running")
        return self._reply

    def set_reply(self, reply: JsonRpcReply) -> None:
        if self.is_ready():
            raise RuntimeError("Task already finished")
        self._reply = reply

    def set_error(self, error: JsonRpcError) -> None:
        if self.is_ready():
            raise RuntimeError("Task already finished")
        self._error = error

    def get_progress(self) -> JsonRpcProgress:
        if not self._progresses:
            raise RuntimeError("No progress received")
        return self._progresses.popleft()

    def add_progress(self, progress: JsonRpcProgress) -> None:
        if self.is_ready():
            raise RuntimeError("Task already finished")
        self._progresses.append(progress)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import asyncio
import threading
from collections.abc import Coroutine
from concurrent.futures import Future
from typing import Any, TypeVar

T = TypeVar("T")


class EventLoop:
    def __init__(self) -> None:
        self._loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self._loop.run_forever, daemon=True)
        self._thread.start()

    @property
    def closed(self) -> bool:
        return self._loop.is_closed()

    def close(self) -> None:
        self._loop.call_soon_threadsafe(self._loop.stop)
        self._thread.join()
        self._loop.close()

    def run(self, coroutine: Coroutine[Any, Any, T]) -> Future[T]:
        return asyncio.run_coroutine_threadsafe(coroutine, self._loop)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
# Responsible Author: Nadir Roman Guerrero <nadir.romanguerrero@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass
from enum import Enum
from typing import Any, ClassVar

from brayns.core import Loader


class ProteinColorScheme(Enum):
    """Define how to color a protein model.

    :param NONE: Doesn't color the proteins.
    :param BY_ID: One color per protein.
    :param PROTEIN_ATOMS: One color per atom.
    :param PROTEIN_CHAINS: One color per protein chain.
    :param PROTEIN_RESIDUES: One color per protein residue.
    """

    NONE = "none"
    BY_ID = "by_id"
    PROTEIN_ATOMS = "protein_atoms"
    PROTEIN_CHAINS = "protein_chains"
    PROTEIN_RESIDUES = "protein_residues"


@dataclass
class ProteinLoader(Loader):
    """Loader for protein files (PDB).

    :param color_scheme: Describe how to color the proteins, defaults to none.
    :type color_scheme: ProteinColorScheme, optional
    :param radius_multiplier: Radius multiplier of the atoms, defaults to 1.
    :type radius_multiplier: float
    """

    PDB: ClassVar[str] = "pdb"
    PDB1: ClassVar[str] = "pdb1"

    color_scheme: ProteinColorScheme = ProteinColorScheme.NONE
    radius_multiplier: float = 1.0

    @classmethod
    @property
    def name(cls) -> str:
        return "protein"

    def get_properties(self) -> dict[str, Any]:
        return {
            "color_scheme": self.color_scheme.value,
            "radius_multiplier": self.radius_multiplier,
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2015-2023, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#                          Nadir Roman <nadir.romanguerrero@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

import setuptools

setuptools.setup()



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Subpackage start braynsService as a subprocess from Python.

Provides mainly helpers to generate the command line.
"""

from .manager import Manager, start
from .process import Process
from .service import LogLevel, Plugin, Service, SslServerContext

__all__ = [
    "LogLevel",
    "Manager",
    "Plugin",
    "Process",
    "Service",
    "SslServerContext",
    "start",
]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import math

from .quaternion import Quaternion
from .vector import Vector3


class Rotation:
    """Arbitrary 3D rotation using normalized quaternion internally."""

    @staticmethod
    def from_quaternion(quaternion: Quaternion) -> Rotation:
        """Construct a rotation using a quaternion (will be normalized).

        :param quaternion: Quaternion representing rotation.
        :type quaternion: Quaternion
        :return: Rotation object.
        :rtype: Rotation
        """
        return Rotation(quaternion)

    @staticmethod
    def from_euler(euler: Vector3, degrees: bool = False) -> Rotation:
        """Construct a rotation from euler angle in XYZ order.

        :param euler: XYZ euler angles.
        :type euler: Vector3
        :param degrees: Euler are provided in degrees, defaults to False.
        :type degrees: bool, optional
        :return: Rotation object.
        :rtype: Rotation
        """
        if degrees:
            euler = Vector3.unpack(math.radians(i) for i in euler)
        quaternion = _euler_to_quaternion(euler)
        return Rotation(quaternion)

    @staticmethod
    def from_axis_angle(axis: Vector3, angle: float, degrees: bool = False) -> Rotation:
        """Construct a rotation of angle around axis.

        Axis must not be zero.

        :param axis: Rotation axis (will be normalized).
        :type axis: Vector3
        :param angle: Rotation angle.
        :type angle: float
        :param degrees: Angle is provided in degrees, defaults to False.
        :type degrees: bool, optional
        :return: Rotation object.
        :rtype: Rotation
        """
        if degrees:
            angle = math.radians(angle)
        quaternion = _axis_angle_to_quaternion(axis, angle)
        return Rotation(quaternion)

    @staticmethod
    def between(source: Vector3, destination: Vector3) -> Rotation:
        """Compute the minimal arc rotation between two vectors.

        Vectors must not be zero.

        :param source: Original direction.
        :type source: Vector3
        :param destination: Final direction.
        :type destination: Vector3
        :return: Rotation object.
        :rtype: Rotation
        """
        quaternion = _get_quaternion_between(source, destination)
        return Rotation.from_quaternion(quaternion)

    @classmethod
    @property
    def identity(cls) -> Rotation:
        """Construct a rotation with no effects.

        :return: Identity rotation.
        :rtype: Rotation
        """
        return Rotation(Quaternion.identity)

    def __init__(self, quaternion: Quaternion) -> None:
        """Direct construction, prefer from_quaternion."""
        self._quaternion = quaternion.normalized

    def __eq__(self, other: object) -> bool:
        """Equality check using internal quaternion."""
        if not isinstance(other, Rotation):
            return False
        return self._quaternion == other.quaternion

    def __hash__(self) -> int:
        """Mark the class as immutable."""
        return hash(self._quaternion)

    @property
    def quaternion(self) -> Quaternion:
        """Get rotation as normalized quaternion.

        :return: Normalized quaternion.
        :rtype: Quaternion
        """
        return self._quaternion

    @property
    def euler_radians(self) -> Vector3:
        """Get rotation as euler angles XYZ in radians.

        :return: Euler angles.
        :rtype: Vector3
        """
        return _quaternion_to_euler(self._quaternion)

    @property
    def euler_degrees(self) -> Vector3:
        """Get rotation as euler angles XYZ in degrees.

        :return: Euler angles.
        :rtype: Vector3
        """
        return Vector3.unpack(math.degrees(i) for i in self.euler_radians)

    @property
    def axis(self) -> Vector3:
        """Get rotation axis.

        :return: Normalized rotation axis.
        :rtype: Vector3
        """
        return self._quaternion.axis

    @property
    def angle_radians(self) -> float:
        """Get rotation angle in radians.

        :return: Rotation angle.
        :rtype: Vector3
        """
        return self._quaternion.angle_radians

    @property
    def angle_degrees(self) -> float:
        """Get rotation angle in degrees.

        :return: Rotation angle.
        :rtype: Vector3
        """
        return self._quaternion.angle_degrees

    @property
    def inverse(self) -> Rotation:
        """Get rotation such as self * self.inverse == Rotation.identity.

        :return: Inverse rotation.
        :rtype: Rotation
        """
        quaternion = self._quaternion.inverse
        return Rotation(quaternion)

    def then(self, other: Rotation) -> Rotation:
        """Combine rotations to be equivalent to self then other.

        :param other: Rotation to apply after self.
        :type other: Rotation
        :return: Rotation object.
        :rtype: Rotation
        """
        quaternion = other._quaternion * self._quaternion
        return Rotation(quaternion)

    def apply(self, value: Vector3, center: Vector3 = Vector3.zero) -> Vector3:
        """Apply rotation on value around center.

        :param value: Vector to rotate.
        :type value: Vector3
        :param center: Rotation center, defaults to Vector3.zero.
        :type center: Vector3, optional
        :return: Rotated value.
        :rtype: Vector3
        """
        quaternion = self._quaternion
        value -= center
        vector = Quaternion(*value)
        vector = quaternion * vector * quaternion.conjugate
        return center + vector.xyz


def euler(x: float, y: float, z: float, degrees: bool = False) -> Rotation:
    """Shortcut to build a rotation from euler angles.

    :param x: X rotation.
    :type x: float
    :param y:  Y rotation.
    :type y: float
    :param z:  Z rotation.
    :type z: float
    :param degrees: Wether given angles are in degrees, defaults to False
    :type degrees: bool, optional
    :return: Rotation corresponding to angles.
    :rtype: Rotation
    """
    angles = Vector3(x, y, z)
    return Rotation.from_euler(angles, degrees)


class ModelRotation:
    """Helper class to store model rotations to reach different views.

    All rotations are relative to front view (X-right, Y-up, Z-front).
    """

    @classmethod
    @property
    def front(cls) -> Rotation:
        return Rotation.identity

    @classmethod
    @property
    def back(cls) -> Rotation:
        return euler(0, 180, 0, degrees=True)

    @classmethod
    @property
    def top(cls) -> Rotation:
        return euler(90, 0, 0, degrees=True)

    @classmethod
    @property
    def bottom(cls) -> Rotation:
        return euler(-90, 0, 0, degrees=True)

    @classmethod
    @property
    def right(cls) -> Rotation:
        return euler(0, -90, 0, degrees=True)

    @classmethod
    @property
    def left(cls) -> Rotation:
        return euler(0, 90, 0, degrees=True)


class CameraRotation:
    """Helper class to store camera rotations to reach different views.

    All rotations are relative to front view (X-right, Y-up, Z-front).
    """

    @classmethod
    @property
    def front(cls) -> Rotation:
        return Rotation.identity

    @classmethod
    @property
    def back(cls) -> Rotation:
        return euler(0, 180, 0, degrees=True)

    @classmethod
    @property
    def top(cls) -> Rotation:
        return euler(-90, 0, 0, degrees=True)

    @classmethod
    @property
    def bottom(cls) -> Rotation:
        return euler(90, 0, 0, degrees=True)

    @classmethod
    @property
    def right(cls) -> Rotation:
        return euler(0, 90, 0, degrees=True)

    @classmethod
    @property
    def left(cls) -> Rotation:
        return euler(0, -90, 0, degrees=True)


def _axis_angle_to_quaternion(axis: Vector3, angle: float) -> Quaternion:
    axis = axis.normalized
    half_angle = angle / 2
    vector = axis * math.sin(half_angle)
    w = math.cos(half_angle)
    return Quaternion(vector.x, vector.y, vector.z, w)


def _euler_to_quaternion(euler: Vector3) -> Quaternion:
    euler /= 2
    cx, cy, cz = Vector3.unpack(math.cos(i) for i in euler)
    sx, sy, sz = Vector3.unpack(math.sin(i) for i in euler)
    return Quaternion(
        sx * cy * cz - cx * sy * sz,
        cx * sy * cz + sx * cy * sz,
        cx * cy * sz - sx * sy * cz,
        cx * cy * cz + sx * sy * sz,
    )


def _quaternion_to_euler(q: Quaternion) -> Vector3:
    return Vector3(_get_x(q), _get_y(q), _get_z(q))


def _get_x(q: Quaternion) -> float:
    sx_cy = 2 * (q.w * q.x + q.y * q.z)
    cx_cy = 1 - 2 * (q.x * q.x + q.y * q.y)
    return math.atan2(sx_cy, cx_cy)


def _get_y(q: Quaternion) -> float:
    sy = 2 * (q.w * q.y - q.z * q.x)
    if abs(sy) >= 1:
        return math.copysign(math.pi / 2, sy)
    return math.asin(sy)


def _get_z(q: Quaternion) -> float:
    sz_cy = 2 * (q.w * q.z + q.x * q.y)
    cz_cy = 1 - 2 * (q.y * q.y + q.z * q.z)
    return math.atan2(sz_cy, cz_cy)


def _get_quaternion_between(u: Vector3, v: Vector3) -> Quaternion:
    angle = _get_angle_between(u, v)
    if angle == 0:
        return Quaternion.identity
    if angle == math.pi:
        axis = _get_orthogonal(u).normalized
        return Quaternion(axis.x, axis.y, axis.z, 0)
    axis = u.cross(v)
    return _axis_angle_to_quaternion(axis, angle)


def _get_angle_between(u: Vector3, v: Vector3) -> float:
    u = u.normalized
    v = v.normalized
    return math.acos(u.dot(v))


def _get_orthogonal(v: Vector3) -> Vector3:
    if v.x != 0:
        x = (v.y + v.z) / v.x
        return Vector3(x, -1, -1)
    if v.y != 0:
        y = (v.x + v.z) / v.y
        return Vector3(-1, y, -1)
    if v.z != 0:
        z = (v.x + v.y) / v.z
        return Vector3(-1, -1, z)
    raise ValueError(v)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import json
from typing import Any

from .json_rpc_error import deserialize_error
from .json_rpc_listener import JsonRpcListener
from .json_rpc_progress import deserialize_progress
from .json_rpc_reply import deserialize_reply, deserialize_reply_from_binary


class JsonRpcDispatcher:
    def __init__(self, listener: JsonRpcListener) -> None:
        self._listener = listener

    def dispatch_binary(self, data: bytes) -> None:
        try:
            reply = deserialize_reply_from_binary(data)
            self._listener.on_reply(reply)
        except Exception as e:
            self._listener.on_invalid_message(e)

    def dispatch_text(self, data: str) -> None:
        try:
            self._dispatch(data)
        except Exception as e:
            self._listener.on_invalid_message(e)

    def _dispatch(self, data: str) -> None:
        message = self._parse(data)
        if self._dispatch_error(message):
            return
        if self._dispatch_reply(message):
            return
        if self._dispatch_progress(message):
            return
        raise ValueError("Unsupported JSON-RPC message")

    def _parse(self, data: str) -> dict[str, Any]:
        message = json.loads(data)
        if not isinstance(message, dict):
            raise ValueError("Message is not a JSON object")
        return message

    def _dispatch_error(self, message: dict[str, Any]) -> bool:
        if "error" not in message:
            return False
        error = deserialize_error(message)
        self._listener.on_error(error)
        return True

    def _dispatch_reply(self, message: dict[str, Any]) -> bool:
        if "result" not in message:
            return False
        reply = deserialize_reply(message)
        self._listener.on_reply(reply)
        return True

    def _dispatch_progress(self, message: dict[str, Any]) -> bool:
        if "id" in message:
            return False
        progress = deserialize_progress(message)
        self._listener.on_progress(progress)
        return True



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from .vector import Vector, Vector3


class PlaneEquation(Vector[float]):
    """Plane equation satisfying ax + by + cz + d = 0.

    Normal and direction point to the upper side of the plane.

    :param a: A component.
    :type a: float
    :param b: B component.
    :type b: float
    :param c: C component.
    :type c: float
    :param d: D component.
    :type d: float
    """

    def __new__(cls, a: float, b: float, c: float, d: float = 0.0) -> PlaneEquation:
        return super().__new__(cls, a, b, c, d)

    @property
    def a(self) -> float:
        return self[0]

    @property
    def b(self) -> float:
        return self[1]

    @property
    def c(self) -> float:
        return self[2]

    @property
    def d(self) -> float:
        return self[3]

    @property
    def normal(self) -> Vector3:
        return Vector3(self.a, self.b, self.c)

    @property
    def direction(self) -> Vector3:
        return self.normal.normalized



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from .json_rpc_dispatcher import JsonRpcDispatcher
from .json_rpc_error import JsonRpcError, deserialize_error
from .json_rpc_future import JsonRpcFuture
from .json_rpc_handler import JsonRpcHandler
from .json_rpc_listener import JsonRpcListener
from .json_rpc_manager import JsonRpcManager
from .json_rpc_progress import JsonRpcProgress, deserialize_progress
from .json_rpc_reply import (
    JsonRpcReply,
    deserialize_reply,
    deserialize_reply_from_binary,
    deserialize_reply_from_text,
)
from .json_rpc_request import (
    JsonRpcRequest,
    serialize_request,
    serialize_request_to_binary,
    serialize_request_to_text,
)
from .json_rpc_task import JsonRpcTask
from .json_rpc_tasks import JsonRpcTasks

__all__ = [
    "deserialize_error",
    "deserialize_progress",
    "deserialize_reply_from_binary",
    "deserialize_reply_from_text",
    "deserialize_reply",
    "JsonRpcDispatcher",
    "JsonRpcError",
    "JsonRpcFuture",
    "JsonRpcHandler",
    "JsonRpcListener",
    "JsonRpcManager",
    "JsonRpcProgress",
    "JsonRpcReply",
    "JsonRpcRequest",
    "JsonRpcTask",
    "JsonRpcTasks",
    "serialize_request_to_binary",
    "serialize_request_to_text",
    "serialize_request",
]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, ClassVar

from brayns.core import Loader

from .morphology import Morphology, serialize_morphology


@dataclass
class SonataNodes:
    """Specify how to load SONATA nodes.

    Use one of the factory methods to create this object.
    """

    density: float | None = None
    names: list[str] | None = None
    ids: list[int] | None = None
    node_count_limit: int | None = None

    def limit(self, node_count_limit: int) -> SonataNodes:
        """Limit the number nodes to be loaded."""
        self.node_count_limit = node_count_limit
        return self

    @staticmethod
    def all() -> SonataNodes:
        """Load all nodes."""
        return SonataNodes.from_density(1.0)

    @staticmethod
    def default() -> SonataNodes:
        """Load a reasonable proportion of nodes."""
        return SonataNodes.from_density(0.01)

    @staticmethod
    def from_density(density: float) -> SonataNodes:
        """Load a proportion of all nodes (0 = none, 1 = all)."""
        return SonataNodes(density=density)

    @staticmethod
    def from_names(names: list[str], density: float = 1.0) -> SonataNodes:
        """Load from node group names with optional density."""
        return SonataNodes(density=density, names=names)

    @staticmethod
    def from_ids(ids: list[int]) -> SonataNodes:
        """Load nodes with given IDs."""
        return SonataNodes(ids=ids)


class SonataReportType(Enum):
    """All available SONATA report types."""

    NONE = "none"
    SPIKES = "spikes"
    COMPARTMENT = "compartment"
    SUMMATION = "summation"
    SYNAPSE = "synapse"
    BLOODFLOW_PRESSURE = "bloodflow_pressure"
    BLOODFLOW_SPEED = "bloodflow_speed"
    BLOODFLOW_RADII = "bloodflow_radii"


@dataclass
class SonataReport:
    """Describe a SONATA report to load.

    Use one of the factory methods to create this object.
    """

    type: SonataReportType
    name: str | None = None
    spike_transition_time: float | None = None

    @staticmethod
    def none() -> SonataReport:
        """No report to load."""
        return SonataReport(
            type=SonataReportType.NONE,
        )

    @staticmethod
    def spikes(spike_transition_time: float = 1.0) -> SonataReport:
        """Spike report with transition time (units from report)."""
        return SonataReport(
            type=SonataReportType.SPIKES, spike_transition_time=spike_transition_time
        )

    @staticmethod
    def compartment(name: str) -> SonataReport:
        """Compartment report with given name."""
        return SonataReport(
            type=SonataReportType.COMPARTMENT,
            name=name,
        )

    @staticmethod
    def summation(name: str) -> SonataReport:
        """Simulation report with given name."""
        return SonataReport(
            type=SonataReportType.SUMMATION,
            name=name,
        )

    @staticmethod
    def synapse(name: str) -> SonataReport:
        """Synapse report with given name."""
        return SonataReport(
            type=SonataReportType.SYNAPSE,
            name=name,
        )

    @staticmethod
    def bloodflow_pressure(name: str) -> SonataReport:
        """Bloodflow pressure report with given name."""
        return SonataReport(
            type=SonataReportType.BLOODFLOW_PRESSURE,
            name=name,
        )

    @staticmethod
    def bloodflow_speed(name: str) -> SonataReport:
        """Bloodflow speed report with given name."""
        return SonataReport(
            type=SonataReportType.BLOODFLOW_SPEED,
            name=name,
        )

    @staticmethod
    def bloodflow_radii(name: str) -> SonataReport:
        """Bloodflow radii report with given name."""
        return SonataReport(
            type=SonataReportType.BLOODFLOW_RADII,
            name=name,
        )


@dataclass
class SonataEdgePopulation:
    """SONATA edge population parameters.

    :param name: Population name.
    :type name: str
    :param afferent: If these edges are afferent or efferent ones.
    :type afferent: bool
    :param afferent: If these edges are afferent ones.
    :type afferent: bool
    :param density: Proportion of edges to load, defaults to 1.
    :type density: float, optional
    :param radius: Edge radii, defaults to 2.
    :type radius: float, optional
    :param report: Report name, defaults to None.
    :type report: str | None, optional
    """

    name: str
    afferent: bool
    density: float = 1.0
    radius: float = 2.0
    report: str | None = None


@dataclass
class SonataNodePopulation:
    """SONATA node population parameters.

    :param name: Population name.
    :type name: str
    :param nodes: Nodes to load, defaults to 0.01% of all nodes.
    :type nodes: SonataNodes, optional
    :param report: Report to load, defaults to None.
    :type report: SonataReport | None, optional
    :param edges: Edges to load, defaults to None.
    :type edges: list[SonataEdgePopulation] | None, optional
    :param morphology: How to load the morphologies, default constructed.
    :type morphology: Morphology, optional
    :param vasculature_radius_multiplier: Vasculature scale, defaults to 1.
    :type vasculature_radius_multiplier: float, optional
    """

    name: str
    nodes: SonataNodes = field(default_factory=SonataNodes.default)
    report: SonataReport | None = None
    edges: list[SonataEdgePopulation] | None = None
    morphology: Morphology = field(default_factory=Morphology)
    vasculature_radius_multiplier: float = 1.0


@dataclass
class SonataLoader(Loader):
    """Loader for SONATA circuit files.

    :param node_populations: Nodes to load.
    :type node_populations: list[SonataNodePopulation]
    """

    JSON: ClassVar[str] = "json"

    node_populations: list[SonataNodePopulation]

    @classmethod
    @property
    def name(cls) -> str:
        return "SONATA loader"

    def get_properties(self) -> dict[str, Any]:
        return _serialize_loader(self)


def _serialize_loader(loader: SonataLoader) -> dict[str, Any]:
    return {
        "node_population_settings": [
            _serialize_node_population(node_population)
            for node_population in loader.node_populations
        ],
    }


def _serialize_node_population(population: SonataNodePopulation) -> dict[str, Any]:
    message: dict[str, Any] = {
        "node_population": population.name,
        "vasculature_geometry_parameters": {
            "radius_multiplier": population.vasculature_radius_multiplier,
        },
        **_serialize_nodes(population.nodes),
        "neuron_morphology_parameters": serialize_morphology(population.morphology),
    }
    if population.report is not None:
        report = _serialize_report(population.report)
        message.update(report)
    if population.edges is not None:
        message["edge_populations"] = [
            _serialize_edge_population(edge) for edge in population.edges
        ]
    return message


def _serialize_nodes(nodes: SonataNodes) -> dict[str, Any]:
    message = dict[str, Any]()
    if nodes.density is not None:
        message["node_percentage"] = nodes.density
    if nodes.names is not None:
        message["node_sets"] = nodes.names
    if nodes.ids is not None:
        message["node_ids"] = nodes.ids
    if nodes.node_count_limit is not None:
        message["node_count_limit"] = nodes.node_count_limit
    return message


def _serialize_report(report: SonataReport) -> dict[str, Any]:
    message: dict[str, Any] = {
        "report_type": report.type.value,
    }
    if report.name is not None:
        message["report_name"] = report.name
    if report.spike_transition_time is not None:
        message["spike_transition_time"] = report.spike_transition_time
    return message


def _serialize_edge_population(population: SonataEdgePopulation) -> dict[str, Any]:
    message: dict[str, Any] = {
        "edge_population": population.name,
        "load_afferent": population.afferent,
        "edge_percentage": population.density,
        "radius": population.radius,
    }
    if population.report is not None:
        message["edge_report_name"] = population.report
    return message



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass
from typing import Any

from brayns.network import Instance, JsonRpcReply
from brayns.utils import ImageFormat, parse_image_format


@dataclass
class ImageInfo:
    """Result of an image rendering with status and encoded data.

    If nothing has been downloaded, data is empty.

    :param accumulation: Current accumulation after render.
    :type accumulation: int
    :param max_accumulation: Accumulation limit to stop rendering.
    :type max_accumulation: int
    :param data: Encoded image data, can be empty.
    :type data: bytes
    """

    accumulation: int
    max_accumulation: int
    data: bytes

    @property
    def full_quality(self) -> bool:
        """Check if max accumulation has been reached.

        :return: True if image is full quality.
        :rtype: bool
        """
        return self.accumulation == self.max_accumulation


@dataclass
class Image:
    """Helper class to take an image of an instance with current settings.

    Note that no render will occur if nothing has changed since last call
    (same context and max accumulation reached).

    If accumulate is True, the instance will render images and accumulate them
    until current renderer ``samples_per_pixel`` is reached. Otherwise only one
    image will be rendered.

    If force_download is True, the framebuffer image will always be downloaded.
    Otherwise it will be downloaded only when something new has been rendered.

    Check ImageInfo.data to see if something has been downloaded.

    :param accumulate: Render all samples at once, defaults to True.
    :type accumulate: bool, optional
    :param force_download: Force download, defaults to True.
    :type force_download: bool, optional
    :param jpeg_quality: JPEG quality.
    :type jpeg_quality: int
    """

    accumulate: bool = True
    force_download: bool = True
    jpeg_quality: int = 100

    def save(self, instance: Instance, path: str, render: bool = True) -> ImageInfo:
        """Try to render an image and save it if rendered.

        :param instance: Instance.
        :type instance: Instance
        :param path: Path to save image.
        :type path: str
        :return: Render status and image data.
        :rtype: ImageInfo
        """
        format = parse_image_format(path)
        image = self.download(instance, format, render)
        if not image.data:
            return image
        with open(path, "wb") as file:
            file.write(image.data)
        return image

    def download(
        self,
        instance: Instance,
        format: ImageFormat = ImageFormat.PNG,
        render: bool = True,
    ) -> ImageInfo:
        """Try to render and download an image.

        :param instance: Instance.
        :type instance: Instance
        :param format: Image encoding format, defaults to ImageFormat.PNG
        :type format: ImageFormat, optional
        :param render: Render image(s) if needed, defaults to True
        :type format: bool, optional
        :return: Render status and image data.
        :rtype: ImageInfo
        """
        params = _serialize_image(self, format=format, render=render)
        return _request(instance, params)

    def render(self, instance: Instance) -> ImageInfo:
        """Try to render an image without downloading it.

        :param instance: Instance.
        :type instance: Instance
        :return: Render status.
        :rtype: ImageInfo
        """
        params = _serialize_image(self, send=False)
        return _request(instance, params)


def _request(instance: Instance, params: dict[str, Any]) -> ImageInfo:
    reply = instance.execute("render-image", params)
    return _deserialize_image(reply)


def _serialize_image(
    image: Image,
    send: bool = True,
    format: ImageFormat = ImageFormat.PNG,
    render: bool = True,
) -> dict[str, Any]:
    params: dict[str, Any] = {
        "send": send,
        "force": send and image.force_download,
        "accumulate": image.accumulate,
        "render": render,
    }
    if send:
        params["format"] = format.value
    if send and format is ImageFormat.JPEG:
        params["jpeg_quality"] = image.jpeg_quality
    return params


def _deserialize_image(reply: JsonRpcReply) -> ImageInfo:
    return ImageInfo(
        accumulation=reply.result["accumulation"],
        max_accumulation=reply.result["max_accumulation"],
        data=reply.binary,
    )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from typing import ClassVar

from brayns.network import Instance
from brayns.utils import Color4


class ColorMethod:
    """Core coloring methods available without plugins."""

    SOLID: ClassVar[str] = "solid"
    GEOMETRY: ClassVar[str] = "primitive"
    TRIANGLE: ClassVar[str] = "triangle"
    VERTEX: ClassVar[str] = "vertex"


def color_model(
    instance: Instance, model_id: int, method: str, colors: dict[str, Color4]
) -> None:
    """Color a model by the given method.

    This function needs a method name and a mapping to get a color from a
    method value.

    See ``ColorMethod`` and ``CircuitColorMethod`` for available methods.

    For example to color a circuit by GID the mapping can be:

    .. code-block:: python

        colors = {
            '1': brayns.Color4.red,
            '2': brayns.Color4.blue,
        }

    Where 1 and 2 are the method values (here GIDs) mapped to the color that
    must be applied to them.

    Supported methods depend on the plugins loaded and the model type. See
    ``get_color_methods`` and ``get_color_values`` for more details.

    When coloring by ID (GIDs, triangle, geometry, ...) the IDs start at zero
    and are incremented in the order of the loading. They can be concatenated
    in a single value using comma and dashes.

    Example: '0,2,4-6,8'.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: ID of the model to color.
    :type model_id: int
    :param method: Coloring method to use.
    :type method: str
    :param colors: Color to use per method value.
    :type colors: dict[str, Color4]
    """
    params = {
        "id": model_id,
        "method": method,
        "values": {key: list(value) for key, value in colors.items()},
    }
    instance.request("color-model", params)


def set_model_color(instance: Instance, model_id: int, color: Color4) -> None:
    """Shortcut to color a model using a single color.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: ID of the model to color.
    :type model_id: int
    :param color: Color to apply on model.
    :type color: Color4
    """
    color_model(instance, model_id, ColorMethod.SOLID, {"color": color})


def get_color_methods(instance: Instance, model_id: int) -> list[str]:
    """Get the available coloring methods for the given model.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: Model ID.
    :type model_id: int
    :return: Color method names.
    :rtype: list[str]
    """
    params = {"id": model_id}
    return instance.request("get-color-methods", params)


def get_color_values(instance: Instance, model_id: int, method: str) -> list[str]:
    """Get available color values for a given method and model.

    Color values are the name of the elements to color. For example, the color
    method 'layer' has values ['1', '2', '3'] if the model has 3 layers.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: Model ID.
    :type model_id: int
    :param method: Coloring method name.
    :type method: str
    :return: Available color values.
    :rtype: list[str]
    """
    params = {
        "id": model_id,
        "method": method,
    }
    return instance.request("get-color-values", params)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class JsonRpcProgress:
    """Request progress info.

    :param id: Request ID.
    :type id: int | str
    :param operation: Description of the current task.
    :type operation: str
    :param amount: Progress amount [0-1].
    :type amount: float
    """

    id: int | str
    operation: str
    amount: float

    def __str__(self) -> str:
        return f"[{self.id}] {self.operation}: {100 * self.amount}%"


def deserialize_progress(message: dict[str, Any]) -> JsonRpcProgress:
    params = message["params"]
    return JsonRpcProgress(
        id=params["id"],
        operation=params["operation"],
        amount=params["amount"],
    )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from brayns.network import Instance
from brayns.utils import Color4, Vector


class ValueRange(Vector[float]):
    """Specify color ramp value range.

    Usually the simulation values are voltages but it can be anything.

    :param min: Minimal simulation value.
    :type min: float
    :param max: Maximal simulation value.
    :type max: float
    """

    def __new__(cls, min: float, max: float) -> ValueRange:
        return super().__new__(cls, min, max)

    @property
    def min(self) -> float:
        return self[0]

    @property
    def max(self) -> float:
        return self[1]

    @property
    def size(self) -> float:
        """Get the difference between min and max.

        :return: Value range.
        :rtype: float
        """
        return self.max - self.min

    def normalize(self, value: float) -> float:
        """Normalize given value between min and max.

        Result is not clamped between min and max.

        For example a value of 5 with min = 0 and max = 10 gives 0.5.

        :param value: Simulation value.
        :type value: float
        :return: Normalized simulation value.
        :rtype: float
        """
        return (value - self.min) / self.size


@dataclass
class ColorRamp:
    """Color ramp to map simulation values to colors.

    Simulation values below value_range.min have colors[0].

    Simulation values above value_range.max have colors[-1].

    Otherwise, the two closest colors in the list are interpolated linearly.

    :param value_range: Simulation value range (usually voltages).
    :type value_range: ValueRange
    :param colors: List of colors mapped to value range.
    :type colors: list[Color4]
    """

    value_range: ValueRange
    colors: list[Color4]


def get_color_ramp(instance: Instance, model_id: int) -> ColorRamp:
    """Retreive the color ramp of the given model.

    Model must have a color ramp attached (usually report loaded).

    :param instance: Instance.
    :type instance: Instance
    :param model_id: Model ID.
    :type model_id: int
    :return: Color ramp.
    :rtype: ColorRamp
    """
    params = {"id": model_id}
    result = instance.request("get-color-ramp", params)
    return _deserialize_color_ramp(result)


def set_color_ramp(instance: Instance, model_id: int, ramp: ColorRamp) -> None:
    """Set the current color ramp of the given model.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: Model ID.
    :type model_id: int
    :param color_ramp: Color ramp.
    :type color_ramp: ColorRamp
    """
    params = {
        "id": model_id,
        "color_ramp": _serialize_color_ramp(ramp),
    }
    instance.request("set-color-ramp", params)


def _deserialize_color_ramp(message: dict[str, Any]) -> ColorRamp:
    return ColorRamp(
        value_range=ValueRange(*message["range"]),
        colors=[Color4(*color) for color in message["colors"]],
    )


def _serialize_color_ramp(ramp: ColorRamp) -> dict[str, Any]:
    return {
        "range": list(ramp.value_range),
        "colors": [list(color) for color in ramp.colors],
    }



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import copy
from dataclasses import dataclass, field

from brayns.network import Instance
from brayns.utils import Rotation, Vector3

from .projection import (
    PerspectiveProjection,
    Projection,
    get_camera_projection,
    set_camera_projection,
)
from .view import View, get_camera_view, set_camera_view


@dataclass
class Camera:
    """Camera used to render.

    A camera is composed of a ``View`` and a ``Projection``.
    """

    view: View = field(default_factory=lambda: View.front)
    projection: Projection = field(default_factory=PerspectiveProjection)

    @property
    def name(self) -> str:
        """Get the camera name (can be compared with ``get_camera_name``).

        :return: Camera projection name.
        :rtype: str
        """
        return self.projection.name

    @property
    def position(self) -> Vector3:
        """Get camera position.

        :return: Camera position.
        :rtype: Vector3
        """
        return self.view.position

    @position.setter
    def position(self, value: Vector3) -> None:
        """Set camera position.

        :param value: New camera position.
        :type value: Vector3
        """
        self.view.position = value

    @property
    def target(self) -> Vector3:
        """Get camera target.

        :return: Camera target.
        :rtype: Vector3
        """
        return self.view.target

    @target.setter
    def target(self, value: Vector3) -> None:
        """Set camera target.

        :param value: New camera target.
        :type value: Vector3
        """
        self.view.target = value

    @property
    def up(self) -> Vector3:
        """Get camera up direction.

        :return: Camera up.
        :rtype: Vector3
        """
        return self.view.up

    @up.setter
    def up(self, value: Vector3) -> None:
        """Set camera up direction.

        :param value: New camera up.
        :type value: Vector3
        """
        self.view.up = value

    @property
    def vector(self) -> Vector3:
        """Get camera vector (target - position).

        :return: Camera vector.
        :rtype: Vector3
        """
        return self.view.vector

    @property
    def direction(self) -> Vector3:
        """Get camera direction (vector.normalized).

        :return: Normalized camera direction.
        :rtype: Vector3
        """
        return self.view.direction

    @property
    def right(self) -> Vector3:
        """Get the camera right direction (direction x up).

        :return: Camera right direction.
        :rtype: Vector3
        """
        return self.view.right

    @property
    def real_up(self) -> Vector3:
        """Get the camera effective up direction (right x direction).

        :return: Camera up direction.
        :rtype: Vector3
        """
        return self.view.real_up

    @property
    def distance(self) -> float:
        """Get distance between camera position and target.

        :return: Camera distance.
        :rtype: float
        """
        return self.view.distance

    @distance.setter
    def distance(self, value: float) -> None:
        """Move the camera position to be at given distance from target.

        Camera direction remains unchanged.

        :param value: New camera distance.
        :type value: float
        """
        self.view.distance = value

    @property
    def orientation(self) -> Rotation:
        """Get camera orientation compared to the front one.

        :return: Camera orientation.
        :rtype: Rotation
        """
        return self.view.orientation

    def translate(self, translation: Vector3) -> Camera:
        """Translate both camera position and target.

        :param translation: Translation to apply.
        :type translation: Vector3
        :return: New translated camera.
        :rtype: Camera
        """
        return Camera(
            view=self.view.translate(translation),
            projection=copy.deepcopy(self.projection),
        )

    def rotate_around_target(self, rotation: Rotation) -> Camera:
        """Rotate camera view around its target.

        :param rotation: Rotation to apply on camera view.
        :type rotation: Rotation
        :return: New rotated camera.
        :rtype: Camera
        """
        return Camera(
            view=self.view.rotate_around_target(rotation),
            projection=copy.deepcopy(self.projection),
        )


def get_camera(instance: Instance, projection_type: type[Projection]) -> Camera:
    """Shortcut to retreive the current camera of an instance.

    Use get_camera_projection and get_camera_view.

    projection_type must be of the same type as the current projection.

    :param instance: Instance.
    :type instance: Instance
    :param projection_type: Current camera projection type of instance.
    :type projection_type: type[Projection]
    :return: Camera with current view and projection.
    :rtype: Camera
    """
    view = get_camera_view(instance)
    projection = get_camera_projection(instance, projection_type)
    return Camera(view, projection)


def set_camera(instance: Instance, camera: Camera) -> None:
    """Shortcut to update the camera view and projection.

    :param instance: Instance.
    :type instance: Instance
    :param camera: Current camera to use for instance.
    :type camera: Camera
    """
    set_camera_view(instance, camera.view)
    set_camera_projection(instance, camera.projection)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

from .json_rpc_error import JsonRpcError
from .json_rpc_listener import JsonRpcListener
from .json_rpc_progress import JsonRpcProgress
from .json_rpc_reply import JsonRpcReply
from .json_rpc_tasks import JsonRpcTasks


class JsonRpcHandler(JsonRpcListener):
    def __init__(self, tasks: JsonRpcTasks, logger: logging.Logger) -> None:
        self._tasks = tasks
        self._logger = logger

    def on_reply(self, reply: JsonRpcReply) -> None:
        self._logger.info("Reply received: %s.", reply)
        self._logger.debug("Reply result: %s.", reply.result)
        self._logger.debug("Reply binary: %d bytes.", len(reply.binary))
        self._tasks.add_reply(reply)

    def on_error(self, error: JsonRpcError) -> None:
        self._logger.info("Error received: %s.", error)
        self._tasks.add_error(error)

    def on_progress(self, progress: JsonRpcProgress) -> None:
        self._logger.info("Progress received: %s.", progress)
        self._tasks.add_progress(progress)

    def on_invalid_message(self, e: Exception) -> None:
        self._logger.error("Invalid message received: %s.", e)
        error = JsonRpcError.general("Invalid message received")
        self._tasks.add_error(error)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, TypeVar

from brayns.network import Instance
from brayns.utils import Color4

T = TypeVar("T", bound="Renderer")


@dataclass
class Renderer(ABC):
    """Base class for all renderer types.

    Accumulation reduce the aliasing but increase the computation time.

    Ray bounces allow non-emissive objects to light other objects.

    :param samples_per_pixel: Accumulation, defaults to 1.
    :type samples_per_pixel: int, optional.
    :param max_ray_bounces: Ray bounces, defaults to 3.
    :type max_ray_bounces: int, optional.
    :param background_color: Background color, defaults to BBP transparent.
    :type background_color: Color4, optional.
    """

    samples_per_pixel: int = 1
    max_ray_bounces: int = 3
    background_color: Color4 = Color4.bbp_background.transparent

    @classmethod
    @property
    @abstractmethod
    def name(cls) -> str:
        """Get renderer name.

        :return: Renderer name.
        :rtype: str
        """
        pass

    @abstractmethod
    def get_additional_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        pass

    @abstractmethod
    def update_additional_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        pass

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "samples_per_pixel": self.samples_per_pixel,
            "max_ray_bounces": self.max_ray_bounces,
            "background_color": list(self.background_color),
            **self.get_additional_properties(),
        }

    @classmethod
    def from_properties(cls: type[T], message: dict[str, Any]) -> T:
        """Low level API to deserialize from JSON."""
        renderer = cls()
        renderer.update_properties(message)
        return renderer

    def get_properties_with_name(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "name": self.name,
            "params": self.get_properties(),
        }

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.samples_per_pixel = message["samples_per_pixel"]
        self.max_ray_bounces = message["max_ray_bounces"]
        self.background_color = Color4(*message["background_color"])
        self.update_additional_properties(message)


@dataclass
class InteractiveRenderer(Renderer):
    """Default renderer used for fast rendering (streaming, tests).

    :param enable_shadows: Enable shadows, defaults to True.
    :type enable_shadows: bool, optional.
    :param ambient_occlusion_samples: AO samples, defaults to 0.
    :type ambient_occlusion_samples: int, optional.
    """

    enable_shadows: bool = True
    ambient_occlusion_samples: int = 0

    @classmethod
    @property
    def name(cls) -> str:
        """Get renderer name.

        :return: Renderer name.
        :rtype: str
        """
        return "interactive"

    def get_additional_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "enable_shadows": self.enable_shadows,
            "ao_samples": self.ambient_occlusion_samples,
        }

    def update_additional_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.enable_shadows = message["enable_shadows"]
        self.ambient_occlusion_samples = message["ao_samples"]


@dataclass
class ProductionRenderer(Renderer):
    """Production renderer for expensive quality rendering.

    Overrides default parameters for higher quality.

    :param samples_per_pixel: Accumulation, defaults to 128.
    :type samples_per_pixel: int, optional.
    :param max_ray_bounces: Ray bounces, defaults to 7.
    :type max_ray_bounces: int, optional.
    """

    samples_per_pixel: int = 128
    max_ray_bounces: int = 7

    @classmethod
    @property
    def name(cls) -> str:
        """Get renderer name.

        :return: Renderer name.
        :rtype: str
        """
        return "production"

    def get_additional_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {}

    def update_additional_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        pass


def get_renderer_name(instance: Instance) -> str:
    """Retreive the name of the current renderer of an instance.

    :param instance: Instance.
    :type instance: Instance
    :return: Renderer name.
    :rtype: str
    """
    return instance.request("get-renderer-type")


def get_renderer(instance: Instance, renderer_type: type[T]) -> T:
    """Retreive the current renderer from an instance.

    Current renderer and given renderer type must be the same.

    Current renderer can be queried using ``get_rendere_name``.

    :param instance: Instance.
    :type instance: Instance
    :param renderer_type: Type of the current renderer.
    :type renderer_type: type[T]
    :return: Current renderer.
    :rtype: T
    """
    name = renderer_type.name
    result = instance.request(f"get-renderer-{name}")
    return renderer_type.from_properties(result)


def set_renderer(instance: Instance, renderer: Renderer) -> None:
    """Set the current renderer of an instance.

    :param instance: Instance.
    :type instance: Instance
    :param renderer: New renderer.
    :type renderer: Renderer
    """
    name = renderer.name
    params = renderer.get_properties()
    instance.request(f"set-renderer-{name}", params)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from typing import Protocol

from .json_rpc_error import JsonRpcError
from .json_rpc_progress import JsonRpcProgress
from .json_rpc_reply import JsonRpcReply


class JsonRpcListener(Protocol):
    def on_reply(self, reply: JsonRpcReply) -> None:
        pass

    def on_error(self, error: JsonRpcError) -> None:
        pass

    def on_progress(self, progress: JsonRpcProgress) -> None:
        pass

    def on_invalid_message(self, e: Exception) -> None:
        pass
